"""Init file for molara package."""

from __future__ import annotations

__copyright__ = "Copyright 2024, Molara"
__version__ = "0.0.2"
"""This variable is used as a single source of truth for the version.

All other parts that need the version such as pyproject.toml and the sphinx
documentation should import it from here.
"""



"""Contains functions to set up vertex attribute objects and vertex buffer objects."""

from __future__ import annotations

import ctypes
from typing import TYPE_CHECKING

import numpy as np
from OpenGL.GL import (
    GL_ARRAY_BUFFER,
    GL_DYNAMIC_DRAW,
    GL_ELEMENT_ARRAY_BUFFER,
    GL_FALSE,
    GL_FLOAT,
    GL_STATIC_DRAW,
    GL_UNSIGNED_INT,
    glBindBuffer,
    glBindVertexArray,
    glBufferData,
    glEnableVertexAttribArray,
    glGenBuffers,
    glGenVertexArrays,
    glVertexAttribDivisor,
    glVertexAttribIPointer,
    glVertexAttribPointer,
)

if TYPE_CHECKING:
    import numpy as np

__copyright__ = "Copyright 2024, Molara"


def setup_vao(
    vertices: np.ndarray,
    indices: np.ndarray,
    model_matrices: np.ndarray,
    colors: np.ndarray,
) -> tuple[int, list[int]]:
    """Set up a vertex attribute object and binds it to the GPU.

    :param vertices: Vertices in the following order x,y,z,r,g,b,nx,ny,nz,..., where xyz are the cartesian coordinates,
        rgb are the color values [0,1], and nxnynz are the components of the normal vector.
    :param indices: Gives the connectivity of the vertices.
    :param model_matrices: Each matrix gives the transformation from object space to world.
    :param colors: Colors of the vertices.
    :return: Returns a bound vertex attribute object
    """
    vao = glGenVertexArrays(1)
    vbo = glGenBuffers(1)

    glBindVertexArray(vao)

    glBindBuffer(GL_ARRAY_BUFFER, vbo)
    glBufferData(GL_ARRAY_BUFFER, vertices.nbytes, vertices, GL_STATIC_DRAW)

    # Vertex positions
    glEnableVertexAttribArray(0)
    glVertexAttribPointer(
        0,
        3,
        GL_FLOAT,
        GL_FALSE,
        vertices.itemsize * 6,
        ctypes.c_void_p(0),
    )

    glEnableVertexAttribArray(1)
    glVertexAttribPointer(
        1,
        3,
        GL_FLOAT,
        GL_FALSE,
        vertices.itemsize * 6,
        ctypes.c_void_p(12),
    )

    ebo = glGenBuffers(1)
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo)
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.nbytes, indices, GL_STATIC_DRAW)

    # Instance colors
    instance_vbo_color = glGenBuffers(1)
    num_instances = len(colors)
    glBindBuffer(GL_ARRAY_BUFFER, instance_vbo_color)
    glBufferData(
        GL_ARRAY_BUFFER,
        num_instances * 3 * colors.itemsize,
        colors,
        GL_DYNAMIC_DRAW,
    )
    glEnableVertexAttribArray(2)
    glVertexAttribPointer(
        2,
        3,
        GL_FLOAT,
        GL_FALSE,
        colors.itemsize * 3,
        ctypes.c_void_p(0),
    )
    glVertexAttribDivisor(2, 1)

    # Instance matrices
    instance_vbo_model = glGenBuffers(1)
    num_instances = len(model_matrices)
    glBindBuffer(GL_ARRAY_BUFFER, instance_vbo_model)
    glBufferData(
        GL_ARRAY_BUFFER,
        num_instances * 16 * model_matrices.itemsize,
        model_matrices,
        GL_DYNAMIC_DRAW,
    )

    for i in range(4):
        glEnableVertexAttribArray(3 + i)
        glVertexAttribPointer(
            3 + i,
            4,
            GL_FLOAT,
            GL_FALSE,
            16 * 4,
            ctypes.c_void_p(i * 16),
        )
        glVertexAttribDivisor(3 + i, 1)
    buffers = [vbo, ebo, instance_vbo_color, instance_vbo_model]
    glBindVertexArray(0)

    return vao, buffers


def setup_vao_numbers(digits: np.ndarray, positions_3d: np.ndarray) -> tuple[int, list[int]]:
    """Set up a vertex attribute object and binds it to the GPU.

    :param digits: The digits to be displayed.
    :param positions_3d: The 3D positions of the atoms.
    :return: Returns a bound vertex attribute object
    """
    # Generate and bind VAO
    vao = glGenVertexArrays(1)
    glBindVertexArray(vao)

    # Generate and bind VBO for instance data
    instance_vbo_digits = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, instance_vbo_digits)
    glBufferData(GL_ARRAY_BUFFER, digits, GL_DYNAMIC_DRAW)
    glEnableVertexAttribArray(0)
    glVertexAttribIPointer(0, 1, GL_UNSIGNED_INT, digits.itemsize, ctypes.c_void_p(0))

    # Generate and bind VBO for instance data
    instance_vbo_positions_3d = glGenBuffers(1)
    glBindBuffer(GL_ARRAY_BUFFER, instance_vbo_positions_3d)
    glBufferData(GL_ARRAY_BUFFER, positions_3d, GL_DYNAMIC_DRAW)
    glEnableVertexAttribArray(1)
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, instance_vbo_positions_3d.itemsize * 3, ctypes.c_void_p(0))

    buffers = [instance_vbo_digits, instance_vbo_positions_3d]
    glBindVertexArray(0)

    return vao, buffers



"""Dialog for specifying a crystal structure."""

from __future__ import annotations

import numpy as np
from PySide6.QtCore import SIGNAL
from PySide6.QtWidgets import QDialog, QMainWindow, QTableWidgetItem

from molara.gui.ui_crystalstructure_dialog import Ui_CrystalDialog
from molara.structure.atom import element_symbol_to_atomic_number
from molara.structure.crystal import Crystal

RIGHTANGLE = 90.0
ENABLED, DISABLED = True, False

__copyright__ = "Copyright 2024, Molara"


class CrystalDialog(QDialog):
    """Dialog for specifying a crystal structure.

    Element symbols, coordinates, lattice constants, supercell size given by user,
    object of type Crystal is instantiated and passed to main window"s OpenGL widget for rendering.
    """

    def __init__(self, parent: QMainWindow | None = None) -> None:
        """Create a CrystalDialog object.

        :param parent: parent widget (main window)
        """
        super().__init__(
            parent,
        )  # main window widget is passed as a parent, so dialog is closed if main window is closed.
        self.ui = Ui_CrystalDialog()
        self.ui.setupUi(self)
        self.list_of_coordinates: list = []
        self.list_of_atomic_numbers: list[int] = []
        self.change_crystal_system("Cubic")
        self.ui.selectCrystalSystem.currentTextChanged.connect(
            self.change_crystal_system,
        )
        self.ui.buttonAddAtom.clicked.connect(self.add_atom)
        self.ui.pushButton.clicked.connect(self.reset)
        self.ui.listAtoms.setColumnCount(4)
        select_space_group = self.ui.selectSpaceGroup
        select_space_group.setCurrentIndex(0)
        self.hide_space_groups([False, True, True, True])

    def reset(self) -> None:
        """Reset the dialog."""
        self.list_of_atomic_numbers = []
        self.list_of_coordinates = []
        self.ui.listAtoms.setRowCount(0)

    def add_atom(self) -> None:
        """Add an atom to the list of atoms."""
        element_symbol = self.ui.inputElementSymbol.text()
        atomic_number = element_symbol_to_atomic_number(element_symbol)
        coord_a, coord_b, coord_c = (
            self.ui.inputAtomCoord_a.value(),
            self.ui.inputAtomCoord_b.value(),
            self.ui.inputAtomCoord_c.value(),
        )
        self.list_of_coordinates += [[coord_a, coord_b, coord_c]]
        self.list_of_atomic_numbers += [atomic_number]
        row_id = len(self.list_of_atomic_numbers) - 1
        self.ui.listAtoms.setRowCount(row_id + 1)
        item_element_symbol = QTableWidgetItem(element_symbol)
        item_coord_a = QTableWidgetItem(str(coord_a))
        item_coord_b = QTableWidgetItem(str(coord_b))
        item_coord_c = QTableWidgetItem(str(coord_c))
        self.ui.listAtoms.setItem(row_id, 0, item_element_symbol)
        self.ui.listAtoms.setItem(row_id, 1, item_coord_a)
        self.ui.listAtoms.setItem(row_id, 2, item_coord_b)
        self.ui.listAtoms.setItem(row_id, 3, item_coord_c)

    def accept(self) -> None:
        """Accept the dialog and passes the crystal to the main window."""
        dim_a, dim_b, dim_c = (
            self.ui.inputSupercell_a.value(),
            self.ui.inputSupercell_b.value(),
            self.ui.inputSupercell_c.value(),
        )
        supercell_dims = [dim_a, dim_b, dim_c]
        a, b, c = (
            self.ui.inputLatConst_a.value(),
            self.ui.inputLatConst_b.value(),
            self.ui.inputLatConst_c.value(),
        )
        alpha, beta, gamma = (
            self.ui.inputLatAngle_alpha.value(),
            self.ui.inputLatAngle_beta.value(),
            self.ui.inputLatAngle_gamma.value(),
        )
        if alpha == RIGHTANGLE and beta == RIGHTANGLE and gamma == RIGHTANGLE:
            basis_vectors = np.diag([a, b, c]).tolist()
        else:
            cosalpha = np.cos(alpha / 180.0 * np.pi)
            cosbeta = np.cos(beta / 180.0 * np.pi)
            cosgamma = np.cos(gamma / 180.0 * np.pi)
            avec = [a, 0.0, 0.0]
            bvec = [
                b * cosgamma,
                b * np.sqrt(1.0 - cosgamma**2),
                0.0,
            ]
            cvec = [
                c * cosbeta,
                c * (cosalpha - cosbeta * cosgamma) / np.sqrt(1.0 - cosgamma**2),
                c
                * np.sqrt(
                    (1.0 - cosalpha**2 - cosbeta**2 - cosgamma**2 + 2 * cosalpha * cosbeta * cosgamma)
                    / (1.0 - cosgamma**2),
                ),
            ]
            basis_vectors = np.array([avec, bvec, cvec]).tolist()
        mycrystal = Crystal(
            self.list_of_atomic_numbers,
            self.list_of_coordinates,
            basis_vectors=basis_vectors,
            supercell_dims=supercell_dims,
        )
        self.parent().ui.openGLWidget.set_structure([mycrystal])  # type: ignore[attr-defined]

    def bc_equals_a(self, value: float) -> None:
        """Set b and c lattice constants equal to a.

        :param value: value of lattice constant 'a'
        """
        self.ui.inputLatConst_b.setValue(value)
        self.ui.inputLatConst_c.setValue(value)

    def b_equals_a(self, value: float) -> None:
        """Set b lattice constant equal to a.

        :param value: value of lattice constant 'a'
        """
        self.ui.inputLatConst_b.setValue(value)

    def angles_hexagonal(self) -> None:
        """Set lattice angles to 90°, 90°, and 120° for a hexagonal cell."""
        self.ui.inputLatAngle_alpha.setValue(90.0)
        self.ui.inputLatAngle_beta.setValue(90.0)
        self.ui.inputLatAngle_gamma.setValue(120.0)
        self.ui.inputLatAngle_alpha.setEnabled(DISABLED)
        self.ui.inputLatAngle_beta.setEnabled(DISABLED)
        self.ui.inputLatAngle_gamma.setEnabled(DISABLED)

    def angles_monoclinic(self) -> None:
        """Set lattice angles to 90°, 90°, and <arbitrary> for a monoclinic cell."""
        self.ui.inputLatAngle_alpha.setValue(90.0)
        self.ui.inputLatAngle_gamma.setValue(90.0)
        self.ui.inputLatAngle_alpha.setEnabled(DISABLED)
        self.ui.inputLatAngle_beta.setEnabled(ENABLED)
        self.ui.inputLatAngle_gamma.setEnabled(DISABLED)

    def angles_orthorhombic(self) -> None:
        """Set lattice angles to 90°, 90°, and 90° for an orthorhombic cell."""
        self.ui.inputLatAngle_alpha.setValue(90.0)
        self.ui.inputLatAngle_beta.setValue(90.0)
        self.ui.inputLatAngle_gamma.setValue(90.0)
        self.ui.inputLatAngle_alpha.setEnabled(DISABLED)
        self.ui.inputLatAngle_beta.setEnabled(DISABLED)
        self.ui.inputLatAngle_gamma.setEnabled(DISABLED)

    def angles_triclinic(self) -> None:
        """Enable lattice inputs for a triclinic cell."""
        self.ui.inputLatAngle_alpha.setEnabled(ENABLED)
        self.ui.inputLatAngle_beta.setEnabled(ENABLED)
        self.ui.inputLatAngle_gamma.setEnabled(ENABLED)

    def enable_lattice_constants(self, ids: list[int]) -> None:
        """Enable or disable inputs for lattice constants, depending on crystal system.

        :param ids: list that contains ids of inputs that shall be enabled
        """
        aid, bid, cid = 0, 1, 2
        self.ui.inputLatConst_a.setEnabled(aid in ids)
        self.ui.inputLatConst_b.setEnabled(bid in ids)
        self.ui.inputLatConst_c.setEnabled(cid in ids)
        receivers_count = self.ui.inputLatConst_a.receivers(
            SIGNAL("valueChanged(double)"),
        )
        if receivers_count > 0:
            self.ui.inputLatConst_a.valueChanged.disconnect()
        if cid in ids and bid not in ids:
            self.ui.inputLatConst_a.valueChanged.connect(self.b_equals_a)
            self.b_equals_a(self.ui.inputLatConst_a.value())
        elif cid not in ids and bid not in ids:
            self.ui.inputLatConst_a.valueChanged.connect(self.bc_equals_a)
            self.bc_equals_a(self.ui.inputLatConst_a.value())

    def hide_space_groups(self, hide: list[bool]) -> None:
        """Hide space-group entries depending on crystal system.

        :param hide: list of bools that specify which space-group entries should be hidden
        """
        view = self.ui.selectSpaceGroup.view()
        for i, hide_i in enumerate(hide):
            view.setRowHidden(i, hide_i)

    def change_crystal_system(self, value: str) -> None:
        """Change the crystal system.

        :param value: name of the crystal system
        """
        self.crystal_system = value
        if value == "Cubic":
            # self.hide_space_groups([False, False, True, True])
            self.enable_lattice_constants([0])
            self.angles_orthorhombic()
        elif value == "Tetragonal":
            # self.hide_space_groups([False, True, False, True])
            self.enable_lattice_constants([0, 2])
            self.angles_orthorhombic()
        elif value == "Orthorhombic":
            # self.hide_space_groups([False, True, True, False])
            self.enable_lattice_constants([0, 1, 2])
            self.angles_orthorhombic()
        elif value == "Hexagonal":
            # self.hide_space_groups([False, True, True, True])
            self.enable_lattice_constants([0, 2])
            self.angles_hexagonal()
        elif value == "Monoclinic":
            # self.hide_space_groups([False, True, True, True])
            self.enable_lattice_constants([0, 1, 2])
            self.angles_monoclinic()
        elif value == "Triclinic":
            # self.hide_space_groups([False, True, True, True])
            self.enable_lattice_constants([0, 1, 2])
            self.angles_triclinic()



"""An importer class for all read in functions."""

from __future__ import annotations

import re
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from molara.structure.atom import element_symbol_to_atomic_number
from molara.structure.crystal import Crystal
from molara.structure.crystals import Crystals
from molara.structure.io.exceptions import FileFormatError

if TYPE_CHECKING:
    from os import PathLike


def robust_split(text: str) -> list[str]:
    """Split a text into its components (separated by any kinds of space characters) with regular expressions.

    :param text: text to be split up
    """
    return re.split(r"\s+", text)


class Importer(ABC):
    """Base class for all importers."""

    def __init__(self, path: PathLike | str) -> None:
        """Initialize the Importer object.

        :param path: input file path
        """
        super().__init__()

        self.path = Path(path)

    @abstractmethod
    def load(self) -> Crystals:
        """Read the file in self.path."""


class PymatgenImporter(Importer):
    """import crystal files.

    This class is a wrapper around the pymatgen Structure class. Supported file formats include cif, poscar, cssr,
    pymatgen's json format, and pymatgen's yaml format.
    """

    def __init__(self, path: PathLike | str) -> None:
        """Initialize the Importer object.

        :param path: input file path
        """
        super().__init__(path)

    def load(self) -> Crystals:
        """Import a file and returns the Crystal."""
        try:
            from pymatgen.core import Structure

            structure = Structure.from_file(self.path)
            crystal = Crystal.from_pymatgen(structure, supercell_dims=[1, 1, 1])
        except ImportError as err:
            msg = "pymatgen is not installed and internal importer not successful, cannot read files"
            raise ImportError(msg) from err
        crystals = Crystals()
        crystals.add_crystal(crystal)
        return crystals


class PoscarImporter(Importer):
    """import crystal files.

    This class can be used to import poscar files. It tries the pymatgen import first.
    """

    def __init__(
        self,
        path: PathLike | str,
    ) -> None:
        """Initialize the Importer object.

        :param path: input file path
        :param supercell_dims: side lengths of the supercell in terms of the cell constants
        """
        super().__init__(path)

    def load(self, use_pymatgen: bool = True) -> Crystals:
        """Import a file and return the Crystal.

        :param use_pymatgen: specifies whether pymatgen (if available) shall be preferred over own parser
        """
        if use_pymatgen:
            try:
                from monty.io import zopen
                from pymatgen.core import Structure as PymatgenStructure

                with zopen(self.path, "rt", errors="replace") as f:
                    contents = f.read()
                structure = PymatgenStructure.from_str(contents, fmt="poscar")
                crystal = Crystal.from_pymatgen(structure, supercell_dims=[1, 1, 1])
            except ImportError:
                warnings.warn("pymatgen is not installed, using internal parser", stacklevel=2)
                crystal = self.parse_poscar()

        else:
            crystal = self.parse_poscar()

        crystals = Crystals()
        crystals.add_crystal(crystal)
        return crystals

    def parse_poscar(self) -> Crystal:
        """Parse POSCAR file (self.path) and return Crystal object."""
        with open(self.path) as file:
            lines = [line.strip() for line in file]
        header_length = 8
        if not len(lines) > header_length:
            msg = "Error: faulty formatting of the POSCAR file."
            raise FileFormatError(msg)
        scale_, latvecs_ = lines[1], lines[2:5]
        species_, numbers_, mode = lines[5:8]
        positions_ = lines[8:]
        try:
            scale = float(scale_)
            basis_vectors = [[float(component) for component in robust_split(latvec_)[:3]] for latvec_ in latvecs_]
            species = robust_split(species_)
            numbers = [int(num) for num in robust_split(numbers_)]
            if len(positions_) == sum(numbers) * 2 + 1:
                positions_ = positions_[0 : sum(numbers)]
            positions = [[float(component) for component in robust_split(pos)[:3]] for pos in positions_]
        except ValueError as err:
            msg = "Error: faulty formatting of the POSCAR file."
            raise FileFormatError(msg) from err
        if (
            len(numbers) != len(species)
            or len(positions) != sum(numbers)
            or not mode.lower().startswith(("d", "c", "k"))  # Either cartesian or direct coords
        ):
            msg = "Error: faulty formatting of the POSCAR file."
            raise FileFormatError(msg)

        # For cartesian coordinates, convert to fractional coordinates
        if mode.lower().startswith(("c", "k")):
            positions = [np.dot(np.linalg.inv(basis_vectors).T, position).tolist() for position in positions]
        atomic_numbers = [element_symbol_to_atomic_number(symb) for symb in species]

        atomic_numbers_extended = []
        for num, an in zip(numbers, atomic_numbers):
            atomic_numbers_extended.extend(num * [an])

        scale_unitcell_to_volume = scale < 0
        if scale_unitcell_to_volume:
            _old_volume = Crystal.calc_volume_unitcell(basis_vectors)
            scale = np.cbrt((-scale) / _old_volume)  # cbrt: cube root

        return Crystal(
            atomic_numbers_extended,
            positions,
            (scale * np.array(basis_vectors)).tolist(),
            supercell_dims=[1, 1, 1],
        )


class VasprunImporter(Importer):
    """import crystal files."""

    def __init__(self, path: PathLike | str) -> None:
        """Initialize the Importer object.

        :param path: input file path
        """
        super().__init__(path)

    def load(self) -> Crystals:
        """Import a file and returns the Crystal."""
        try:
            from pymatgen.io.vasp import Vasprun

            vasprun = Vasprun(self.path)
            structure = vasprun.final_structure
            crystal = Crystal.from_pymatgen(structure, supercell_dims=[1, 1, 1])
        except ImportError as err:
            msg = "pymatgen is not installed, cannot read vasprun.xml files"
            raise FileFormatError(
                msg,
            ) from err
        crystals = Crystals()
        crystals.add_crystal(crystal)
        return crystals



"""Contains the Drawer class."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from molara.rendering.cylinder import Cylinder, calculate_cylinder_model_matrix
from molara.rendering.matrices import (
    calculate_model_matrices,
    calculate_rotation_matrices,
    calculate_scale_matrices,
    calculate_translation_matrices,
)
from molara.rendering.sphere import (
    Sphere,
)
from molara.tools.mathtools import norm

if TYPE_CHECKING:
    from molara.structure.atom import Atom

__copyright__ = "Copyright 2024, Molara"


NO_BONDS = np.array([[-1, -1]], dtype=np.int_)


class Drawer:
    """Creates a Drawer object."""

    def __init__(self, atoms: list[Atom], bonds: np.ndarray, draw_bonds: bool) -> None:
        """Create a Drawer object.

        :param atoms: list of atoms to be drawn
        :param bonds: list ids of bonded atoms
        :param draw_bonds: bool that specifies whether bonds shall be drawn (as cylinders)
        """
        self.subdivisions_sphere = 20
        self.subdivisions_cylinder = 20
        self.stick_mode = False
        self.color_scheme = "Jmol"

        self.sphere_scale = 1.0
        self.sphere_default_radius = 1.0 / 6
        self.cylinder_scale = 1.0
        self.cylinder_radius = 0.075

        self.sphere = Sphere(self.subdivisions_sphere)
        self.cylinder = Cylinder(self.subdivisions_cylinder)
        self.sphere_model_matrices = np.array([], dtype=np.float32)
        self.sphere_translation_matrices: list | np.ndarray = []
        self.sphere_scale_matrices: list | np.ndarray = []

        self.atom_positions: list | np.ndarray = []
        self.atom_colors: np.ndarray = np.array([], dtype=np.float32)
        self.atom_scales: list | np.ndarray = []
        self.update_atoms(atoms)

        self.cylinder_scale_matrices: list | np.ndarray = []
        self.cylinder_rotation_matrices: list | np.ndarray = []
        self.cylinder_translation_matrices: list | np.ndarray = []
        self.cylinder_model_matrices = np.array([], dtype=np.float32)
        self.cylinder_colors: list | np.ndarray = np.array([], dtype=np.float32)
        self.cylinder_positions: list | np.ndarray = []
        self.cylinder_directions: list | np.ndarray = []
        self.cylinder_dimensions: list | np.ndarray = []

        self.bonds = NO_BONDS
        self.update_bonds(bonds, draw_bonds)

    @property
    def has_bonds(self) -> bool:
        """Specifies whether drawer has been passed any bonds to draw."""
        return self.bonds[0][0] != -1

    def update_atoms(self, atoms: list[Atom] | None = None) -> None:
        """Update the bonds and/or bond matrices of the drawer."""
        if atoms is not None:
            self.atoms = atoms
            self.set_atom_colors()
            self.set_atom_positions()
            self.set_atom_scales()
            self.set_atom_scale_matrices()
        self.set_atom_translation_matrices()
        self.set_atom_model_matrices()

    def update_bonds(self, bonds: np.ndarray | None = None, draw_bonds: bool = True) -> None:
        """Update the bonds and/or bond matrices of the drawer."""
        self.draw_bonds = draw_bonds

        if bonds is not None:
            self.bonds = bonds

        if not self.draw_bonds:
            return

        if self.has_bonds:
            self.set_cylinder_props()
            if bonds is not None:
                self.set_cylinder_scale_matrices()
                self.set_cylinder_rotation_matrices()
            self.set_cylinder_translation_matrices()
            self.set_cylinder_model_matrices()

    def set_atoms(self, atoms: list[Atom]) -> None:
        """Set the atoms to be displayed.

        :param atoms: List of atoms to be displayed.
        """
        self.atoms = atoms
        self.set_atom_colors()
        self.set_atom_positions()

    def set_atom_colors(self) -> None:
        """Set the colors of the atoms."""
        self.atom_colors = np.array([atom.color[self.color_scheme] for atom in self.atoms], dtype=np.float32)

    def set_cylinder_dimensions(self) -> None:
        """Set the dimensions of the cylinders.

        It is important to note, that the radius ([:,1]) of the cylinders need not be changed!
        """
        self.cylinder_dimensions = np.array(self.cylinder_dimensions, dtype=np.float32)
        self.cylinder_dimensions[:, 0] = self.cylinder_radius * self.cylinder_scale
        self.cylinder_dimensions[:, 2] = self.cylinder_radius * self.cylinder_scale

    def set_cylinder_props(self) -> None:
        """Set the colors of the bonds (cylinders)."""
        self.cylinder_colors = []
        self.cylinder_positions = []
        self.cylinder_directions = []
        self.cylinder_dimensions = []
        radius = self.cylinder_radius * self.cylinder_scale

        for bond in self.bonds:
            if bond[0] == -1:
                continue

            atom1, atom2 = self.atoms[bond[0]], self.atoms[bond[1]]
            pos1, pos2 = atom1.position, atom2.position
            difference = pos1 - pos2
            # Calculate the length of the cylinder.
            length = float(norm(difference)) / 2
            mid_point = (pos1 + pos2) / 2
            # calculate the point 1 quarter between the 2 atoms
            pos_quarter1 = mid_point + difference / 4
            # calculate the point 3 quarter between the 2 atoms
            pos_quarter2 = mid_point - difference / 4
            self.cylinder_positions.append(pos_quarter1)
            self.cylinder_positions.append(pos_quarter2)
            self.cylinder_directions.append(difference)
            self.cylinder_dimensions.append([radius, length, radius])
            self.cylinder_colors += [
                np.array([atom1.color[self.color_scheme]], dtype=np.float32),
                np.array([atom2.color[self.color_scheme]], dtype=np.float32),
            ]

        self.cylinder_colors = np.array(self.cylinder_colors, dtype=np.float32)
        self.cylinder_positions = np.array(self.cylinder_positions, dtype=np.float32)
        self.cylinder_directions = np.array(self.cylinder_directions, dtype=np.float64)
        self.cylinder_dimensions = np.array(self.cylinder_dimensions, dtype=np.float32)

    def set_cylinder_colors(self) -> None:
        """Set the colors of the bonds (cylinders)."""
        self.cylinder_colors = []
        for bond in self.bonds:
            if bond[0] == -1:
                continue
            atom1, atom2 = self.atoms[bond[0]], self.atoms[bond[1]]
            self.cylinder_colors += [
                np.array([atom1.color[self.color_scheme]], dtype=np.float32),
                np.array([atom2.color[self.color_scheme]], dtype=np.float32),
            ]
        self.cylinder_colors = np.array(self.cylinder_colors, dtype=np.float32)

    def set_atom_positions(self) -> None:
        """Set the positions of the atoms."""
        self.atom_positions = np.array(
            [np.array(atom.position, dtype=np.float32) for atom in self.atoms],
            dtype=np.float32,
        )

    def set_atom_scales(self) -> None:
        """Set the scales of the atoms."""
        self.atom_scales = []
        scaling_factor = self.sphere_default_radius * self.sphere_scale
        if not self.stick_mode:
            self.atom_scales = np.array(
                [3 * [scaling_factor * atom.vdw_radius] for atom in self.atoms],
                dtype=np.float32,
            )
        else:
            self.atom_scales = np.array([3 * [scaling_factor] for _ in self.atoms], dtype=np.float32)

    def reset_atom_model_matrices(self) -> None:
        """Reset the model matrices for the spheres."""
        self.sphere_model_matrices = np.array([], dtype=np.float32)

    def reset_cylinders_model_matrices(self) -> None:
        """Reset the model matrices for the cylinders."""
        self.cylinder_model_matrices = np.array([], dtype=np.float32)

    def reset_atom_colors(self) -> None:
        """Reset the colors for the spheres."""
        self.atom_colors = np.array([], dtype=np.float32)

    def set_cylinder_model_matrices(self) -> None:
        """Set the model matrices for the cylinders."""
        self.reset_cylinders_model_matrices()
        self.cylinder_model_matrices = calculate_model_matrices(
            self.cylinder_translation_matrices,
            self.cylinder_scale_matrices,
            self.cylinder_rotation_matrices,
            cylinder=True,
        )

    def set_cylinder_translation_matrices(self) -> None:
        """Set the translation matrices for the spheres."""
        self.cylinder_translation_matrices = calculate_translation_matrices(
            np.array(self.cylinder_positions),
        )

    def set_cylinder_scale_matrices(self) -> None:
        """Set the translation matrices for the cylinders."""
        self.cylinder_scale_matrices = calculate_scale_matrices(
            np.array(self.cylinder_dimensions),
        )

    def set_cylinder_rotation_matrices(self) -> None:
        """Set the rotation matrices for the cylinders."""
        self.cylinder_rotation_matrices = calculate_rotation_matrices(
            np.array(self.cylinder_directions),
        )

    def set_atom_translation_matrices(self) -> None:
        """Set the translation matrices for the spheres."""
        self.sphere_translation_matrices = calculate_translation_matrices(
            np.array(self.atom_positions),
        )

    def set_atom_scale_matrices(self) -> None:
        """Set the scale matrices for the spheres."""
        self.sphere_scale_matrices = calculate_scale_matrices(
            np.array(self.atom_scales),
        )

    def set_atom_model_matrices(self) -> None:
        """Set the model matrices for the spheres."""
        self.reset_atom_model_matrices()
        self.sphere_model_matrices = calculate_model_matrices(
            np.array(self.sphere_translation_matrices),
            np.array(self.sphere_scale_matrices),
        )


def calculate_bond_cylinders_model_matrix(atom1: Atom, atom2: Atom) -> np.ndarray:
    """Calculate the model matrix for a cylinder between two atoms.

    :param atom1: first Atom
    :param atom2: second Atom
    :return: Model matrix for the cylinder between atom1 and atom2.
    """
    radius = 0.075
    difference = atom1.position - atom2.position
    # Calculate the length of the cylinder.
    length = float(norm(difference)) / 2
    mid_point = (atom1.position + atom2.position) / 2
    # calculate the point 1 quarter between the 2 atoms
    pos_quarter1 = mid_point - difference / 4
    # calculate the point 3 quarter between the 2 atoms
    pos_quarter2 = mid_point + difference / 4

    mat1 = calculate_cylinder_model_matrix(pos_quarter1, radius, length, difference)
    mat2 = calculate_cylinder_model_matrix(pos_quarter2, radius, length, difference)
    return np.array(
        [
            mat2,
            mat1,
        ],
        dtype=np.float32,
    )



"""Dialog for exporting a snapshot of the rendered structure."""

from __future__ import annotations

import time
from pathlib import Path

from PySide6.QtWidgets import QDialog, QFileDialog, QMainWindow

from molara.gui.ui_export_image_dialog import Ui_Dialog
from molara.gui.virtual_rendering_window import VirtualRenderingWindow


class ExportImageDialog(QDialog):
    """Dialog for exporting a snapshot of the rendered structure."""

    def __init__(self, parent: QMainWindow) -> None:
        """Instantiate the dialog object."""
        super().__init__(parent)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.set_event_connections()

        self.main_window = parent

        # temporary adjustments for the dialog until features are implemented
        self.ui.tabWidget.setTabEnabled(1, False)  # noqa: FBT003 (disable the "Advanced" tab for now)
        self.ui.checkBox.setChecked(True)
        self.ui.checkBox.setDisabled(True)

    def set_event_connections(self) -> None:
        """Connect the dialog events to functions."""
        self.ui.widthSpinBox.valueChanged.connect(self.update_height)
        self.ui.heightSpinBox.valueChanged.connect(self.update_width)
        self.ui.pushButton.clicked.connect(self.choose_filename)

    def choose_filename(self) -> None:
        """Open a file dialog to choose a filename."""
        filename, suffix = QFileDialog.getSaveFileName(
            self,
            "Export Image",
            f"{Path.home()}",
            "*.png;;*.jpg",
        )
        if not filename:
            return
        if not filename.endswith(suffix[1:]):
            filename += suffix[1:]
        self.ui.filenameInput.setText(filename)

    def update_height(self) -> None:
        """Update the height value in the form."""
        width = self.ui.widthSpinBox.value()
        height = self.main_window.structure_widget.height() * width // self.main_window.structure_widget.width()
        self.ui.heightSpinBox.blockSignals(True)  # noqa: FBT003 (block signals to prevent infinite loop)
        self.ui.heightSpinBox.setValue(height)
        self.ui.heightSpinBox.blockSignals(False)  # noqa: FBT003 (block signals to prevent infinite loop)

    def update_width(self) -> None:
        """Update the width value in the form."""
        height = self.ui.heightSpinBox.value()
        width = self.main_window.structure_widget.width() * height // self.main_window.structure_widget.height()
        self.ui.widthSpinBox.blockSignals(True)  # noqa: FBT003 (block signals to prevent infinite loop)
        self.ui.widthSpinBox.setValue(width)
        self.ui.widthSpinBox.blockSignals(False)  # noqa: FBT003 (block signals to prevent infinite loop)

    def show_dialog(self) -> None:
        """Show the dialog and set default values in the form."""
        self.ui.widthSpinBox.setValue(self.main_window.structure_widget.width())
        self.ui.heightSpinBox.setValue(self.main_window.structure_widget.height())
        self.ui.filenameInput.setText(f"{Path.home()}/molara_image_{time.time():1.0f}.png")
        self.show()

    def accept(self) -> None:
        """Export the image and close the dialog."""
        # gather export specifics
        width, height = int(self.ui.widthSpinBox.value()), int(self.ui.heightSpinBox.value())
        virtual_rendering_window = VirtualRenderingWindow(self.main_window)
        virtual_rendering_window.export_image(self.ui.filenameInput.text(), custom_geometry=(width, height))
        del virtual_rendering_window



"""Contains the StructureWidget class, which is a subclass of QOpenGLWidget."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import numpy as np
from OpenGL.GL import GL_DEPTH_TEST, GL_MULTISAMPLE, glClearColor, glEnable, glViewport
from PySide6.QtCore import QEvent, Qt
from PySide6.QtGui import QGuiApplication
from PySide6.QtOpenGLWidgets import QOpenGLWidget

from molara.rendering.atom_labels import calculate_atom_number_arrays
from molara.rendering.camera import Camera
from molara.rendering.rendering import Renderer
from molara.rendering.shaders import compile_shaders
from molara.structure.crystal import Crystal
from molara.tools.raycasting import select_sphere

if TYPE_CHECKING:
    from PySide6.QtGui import QMouseEvent
    from PySide6.QtWidgets import QWidget

    from molara.structure.molecule import Molecule
    from molara.structure.structure import Structure

__copyright__ = "Copyright 2024, Molara"

MEASUREMENT, BUILDER = 0, 1


class StructureWidget(QOpenGLWidget):
    """Creates a StructureWidget object, which is a subclass of QOpenGLWidget."""

    def __init__(self, parent: QWidget) -> None:
        """Create a StructureWidget object, which is a subclass of QOpenGLWidget.

        :param parent: parent widget (main window's central widget)
        """
        self.central_widget = parent
        self.main_window = self.central_widget.parent()  # type: ignore[method-assign, assignment]
        QOpenGLWidget.__init__(self, parent)

        self.renderer = Renderer()
        self.structures: list[Structure | Molecule | Crystal] = []
        self.vertex_attribute_objects = [-1]
        self.axes = [
            -1,
            -1,
        ]  # -1 means no axes are drawn, any other integer means axes are drawn
        self.box = [
            -1,
            -1,
        ]
        self.rotate = False
        self.translate = False
        self.click_position: np.ndarray | None = None
        self.rotation_angle_x = 0.0
        self.rotation_angle_y = 0.0
        self.position = np.zeros(2)
        self.old_position = np.zeros(2)
        self.contour = False
        self.camera = Camera(self.width(), self.height())
        self.cursor_in_widget = False
        self.measurement_selected_spheres: list = [-1] * 4
        self.builder_selected_spheres: list = [-1] * 3

        self.old_sphere_colors: list = [np.ndarray] * 4
        self.new_sphere_colors: list = [
            np.array([1, 0, 0], dtype=np.float32),
            np.array([0, 1, 0], dtype=np.float32),
            np.array([0, 0, 1], dtype=np.float32),
            np.array([1, 1, 0], dtype=np.float32),
        ]
        self.show_atom_indices = False
        self.atom_indices_arrays: tuple[np.ndarray, np.ndarray] = (np.zeros(1), np.zeros(1))
        self.number_scale = 1.0

    @property
    def bonds(self) -> bool:
        """Specifies whether bonds should be drawn (returns False if no bonds present whatsoever)."""
        if self.structures:
            result = self.structures[0].draw_bonds and self.structures[0].has_bonds
            self.main_window.structure_customizer_dialog.bonds = result
            return result
        return False

    @property
    def draw_bonds(self) -> bool:
        """Specifies whether bonds should be drawn."""
        if len(self.structures) != 1:
            return False
        return self.structures[0].draw_bonds

    @property
    def draw_axes(self) -> bool:
        """Specifies whether the axes should be drawn."""
        return self.axes[0] != -1

    @property
    def draw_unit_cell_boundaries(self) -> bool:
        """Specifies whether the unit cell boundaries should be drawn."""
        return self.box[0] != -1

    @property
    def orthographic_projection(self) -> bool:
        """Specifies whether the projection is orthographic or not."""
        return self.camera.orthographic_projection

    def update_atom_number_labels(self) -> None:
        """Update the positions of the labels."""
        assert self.structures

        calculate_atom_number_arrays(
            self.atom_indices_arrays[0],
            self.atom_indices_arrays[1],
            self.structures[0],
            self.camera,
        )
        self.renderer.draw_numbers(
            self.atom_indices_arrays[0],
            self.atom_indices_arrays[1],
        )

    def reset_view(self) -> None:
        """Reset the view of the structure to the initial view."""
        self.center_structure()
        dy, dz = None, None
        if not len(self.structures) or not self.structures[0]:
            return
        if not self.structures[0].atoms:
            return
        if len(self.structures[0].atoms) > 1:
            x, y, z = np.array([atom.position for atom in self.structures[0].atoms]).T
            dy = y.max() - y.min()
            dz = z.max() - z.min()
        self.camera.reset(self.width(), self.height(), dy, dz)
        self.update()

    def set_view_to_axis(self, axis: Literal["x", "y", "z"]) -> None:
        """Set view angle parallel to a specified axis.

        :param axis: axis to which the view shall be set
        """
        self.camera.center_coordinates()
        self.camera.set_rotation(axis)
        self.update()

    def delete_structure(self) -> None:
        """Delete structures and reset vertex attributes."""
        self.vertex_attribute_objects = [-1]
        self.update()

    def set_structure(self, structs: list[Structure | Crystal | Molecule], reset_view: bool = True) -> None:
        """Set the structures to be drawn.

        :param structs: list of Structure object that shall be drawn
        :param reset_view: Specifies whether the view shall be reset to the initial view
        """
        self.structures = structs
        if reset_view:
            self.reset_view()
        else:
            self.set_vertex_attribute_objects()
            self.update()

        self.main_window.structure_customizer_dialog.set_bonds(self.bonds)
        self.main_window.structure_customizer_dialog.apply_changes()
        self.toggle_unit_cell_boundaries(update_box=True)

        self.reset_measurement()

    def center_structure(self) -> None:
        """Centers the structure in the widget."""
        if not self.structures:
            return
        self.structures[0].center_coordinates()
        self.camera.center_coordinates()
        self.set_vertex_attribute_objects()
        self.update()

    def export_camera_settings(self, filename: str) -> None:
        """Export camera settings to .npz file.

        :param filename: name of the file to which the camera settings shall be saved
        """
        self.camera.export_settings(filename)

    def import_camera_settings(self, filename: str) -> None:
        """Import camera settings from .npz file.

        :param filename: name of the file from which the camera settings shall be loaded
        """
        self.camera.import_settings(filename)
        self.update()

    def initializeGL(self) -> None:  # noqa: N802
        """Initialize the widget."""
        glClearColor(1, 1, 1, 1.0)
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_MULTISAMPLE)
        self.renderer.set_shaders(compile_shaders())
        self.renderer.aspect_ratio = self.width() / self.height()

    def resizeGL(self, width: int, height: int) -> None:  # noqa: N802
        """Resizes the widget.

        :param width: widget width (in pixels)
        :param height: widget height (in pixels)
        """
        glViewport(0, 0, width, height)  # one can also use self.width() and self.height()
        self.camera.width, self.camera.height = width, height
        self.camera.calculate_projection_matrix()
        self.renderer.aspect_ratio = self.width() / self.height()
        self.update()

    def paintGL(self) -> None:  # noqa: N802
        """Draws the scene."""
        self.renderer.draw_scene(self.camera, self.bonds)
        if self.show_atom_indices:
            self.update_atom_number_labels()
            self.renderer.display_numbers(self.camera, self.number_scale)

    def set_vertex_attribute_objects(self, update_bonds: bool = True) -> None:
        """Set the vertex attribute objects of the structure."""
        assert isinstance(self.structures[0].drawer.cylinder_colors, np.ndarray)
        sphere_vertices = self.structures[0].drawer.sphere.vertices
        sphere_indices = self.structures[0].drawer.sphere.indices
        cylinder_vertices = self.structures[0].drawer.cylinder.vertices
        cylinder_indices = self.structures[0].drawer.cylinder.indices
        sphere_model_matrices = self.structures[0].drawer.sphere_model_matrices
        atom_colors = self.structures[0].drawer.atom_colors
        cylinder_model_matrices = self.structures[0].drawer.cylinder_model_matrices
        cylinder_colors = self.structures[0].drawer.cylinder_colors
        for i in range(1, len(self.structures)):
            sphere_model_matrices = np.concatenate(
                (sphere_model_matrices, self.structures[i].drawer.sphere_model_matrices),
                axis=0,
            )
            atom_colors = np.concatenate(
                (atom_colors, self.structures[i].drawer.atom_colors),
                axis=0,
            )
            cylinder_model_matrices = np.concatenate(
                (cylinder_model_matrices, self.structures[i].drawer.cylinder_model_matrices),
                axis=0,
            )
            cylinder_colors = np.concatenate(
                (cylinder_colors, self.structures[i].drawer.cylinder_colors),
                axis=0,
            )
        self.makeCurrent()
        self.renderer.update_atoms_vao(
            sphere_vertices,
            sphere_indices,
            sphere_model_matrices,
            atom_colors,
        )
        self.renderer.update_bonds_vao(
            cylinder_vertices,
            cylinder_indices,
            cylinder_model_matrices,
            cylinder_colors,
        ) if update_bonds else None

    def wheelEvent(self, event: QEvent) -> None:  # noqa: N802
        """Zooms in and out of the structure."""
        num_degrees = event.angleDelta().y() / 8  # type: ignore[attr-defined]
        num_steps = num_degrees / 100  # Empirical value to control zoom speed
        self.camera.set_distance_from_target(num_steps)
        self.camera.update()
        self.update()

    def mousePressEvent(self, event: QMouseEvent) -> None:  # noqa: N802
        """Start the rotation or translation of the structure.

        :param event: mouse event (such as left click, right click...)
        """
        if event.position().x() not in range(self.width()) or event.position().y() not in range(self.height()):
            return

        if event.button() == Qt.MouseButton.LeftButton:
            # first test if Shift key is pressed (for selecting atoms)
            if bool(QGuiApplication.keyboardModifiers() & Qt.ShiftModifier):  # type: ignore[attr-defined]
                if self.main_window.measurement_dialog.isVisible():
                    self.update_selected_atoms(MEASUREMENT, event)
                if self.main_window.builder_dialog.isVisible():
                    self.update_selected_atoms(BUILDER, event)
                return

            self.rotate = True
            if self.translate is True:
                self.stop_translate(event)
            self.set_normalized_position(event)
            self.click_position = np.copy(self.position)

        if event.button() == Qt.MouseButton.RightButton:
            self.translate = True
            if self.rotate is True:
                self.stop_rotation(event)
            self.set_normalized_position(event)
            self.click_position = np.copy(self.position)

    def mouseMoveEvent(self, event: QMouseEvent) -> None:  # noqa: N802
        """Rotates or translates the structure.

        :param event: mouse event (such as left click, right click...)
        """
        if self.rotate and self.click_position is not None:
            self.set_normalized_position(event)
            self.camera.set_rotation_quaternion(self.click_position, self.position)
        if self.translate and self.click_position is not None:
            self.set_normalized_position(event)
            self.camera.set_translation_vector(self.click_position, self.position)
        self.camera.update()
        self.update()

    def set_normalized_position(self, event: QMouseEvent) -> None:
        """Set the normalized position of the mouse cursor.

        :param event: mouse event (such as left click, right click...)
        """
        if self.width() >= self.height():
            self.position[0] = (event.position().x() * 2 - self.width()) / self.width()
            self.position[1] = -(event.position().y() * 2 - self.height()) / self.width()
        else:
            self.position[0] = (event.position().x() * 2 - self.width()) / self.height()
            self.position[1] = -(event.position().y() * 2 - self.height()) / self.height()
        self.position = np.array(self.position, dtype=np.float32)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:  # noqa: N802
        """Stop the rotation or translation of the structure.

        :param event: mouse event (such as left click, right click...)
        """
        if event.button() == Qt.MouseButton.LeftButton and self.rotate:
            self.stop_rotation(event)
        if event.button() == Qt.MouseButton.RightButton and self.translate:
            self.stop_translate(event)

    def stop_translate(self, event: QMouseEvent) -> None:
        """Stop the translation of the structure.

        :param event: mouse event (such as left click, right click...)
        :return:
        """
        self.translate = False
        self.set_normalized_position(event)
        self.camera.update(save=True)
        self.click_position = None

    def stop_rotation(self, event: QMouseEvent) -> None:
        """Stop the rotation of the structure.

        :param event: mouse event (such as left click, right click...)
        :return:
        """
        self.rotate = False
        self.set_normalized_position(event)
        self.camera.update(save=True)
        self.click_position = None

    def toggle_axes(self) -> None:
        """Draws the cartesian axes."""
        length = 2.0
        radius = 0.02
        self.makeCurrent()
        if self.draw_axes:
            self.renderer.remove_cylinder(self.axes[0])
            self.renderer.remove_sphere(self.axes[1])
            self.axes = [-1, -1]
            self.update()
            self.main_window.update_action_texts()
            return

        positions = np.array(
            [[length / 2, 0, 0], [0, length / 2, 0], [0, 0, length / 2]],
            dtype=np.float32,
        )
        directions = np.eye(3, dtype=np.float32)
        colors = np.eye(3, dtype=np.float32)
        radii = np.array([radius] * 3, dtype=np.float32)
        lengths = np.array([length] * 3, dtype=np.float32)
        self.axes[0] = self.renderer.draw_cylinders(
            positions,
            directions,
            radii,
            lengths,
            colors,
            25,
        )
        positions = np.array(
            [[length, 0, 0], [0, length, 0], [0, 0, length], [0, 0, 0]],
            dtype=np.float32,
        )
        colors = np.array(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 1]],
            dtype=np.float32,
        )
        radii = np.array([radius] * 4, dtype=np.float32)
        self.axes[1] = self.renderer.draw_spheres(positions, radii, colors, 25)
        self.update()

        self.main_window.update_action_texts()

    def toggle_projection(self) -> None:
        """Toggles between orthographic and perspective projection."""
        self.camera.toggle_projection()
        self.update()
        self.main_window.update_action_texts()

    def toggle_unit_cell_boundaries(self, update_box: bool = False) -> None:
        """Draws the unit cell boundaries.

        :param update_box: specifies whether box shall be updated. If False, a drawn box will be hidden.
        """
        if len(self.structures) != 1:
            return

        self.makeCurrent()

        box_was_drawn = self.box[0] != -1

        if not box_was_drawn and update_box:
            # if no box is drawn and unit cell boundary shall not be toggled but just updated, nothing needs to be done!
            return

        if box_was_drawn:
            self.renderer.remove_cylinder(self.box[0])
            # if a box was drawn and the unit cell boundary shall not be simply updated,
            # the box should be removed.
            if not update_box:
                self.box = [-1, -1]
                self.update()
                self.main_window.update_action_texts()
                return
            if not isinstance(self.structures[0], Crystal):
                self.box = [-1, -1]
                self.update()
                self.main_window.update_action_texts()
                return

        # the unit cell boundaries shall be drawn anew if:
        # 1.) a box was not drawn before and function is called as a "toggle", not an update
        # 2.) a box was drawn before, but shall be updated (crystal structure changed)
        assert isinstance(self.structures[0], Crystal)

        positions = self.structures[0].unitcell_boundaries_positions

        diagonal_length = np.linalg.norm(self.structures[0].basis_vectors)
        lowerlim_radius = 0.005
        radius = max(lowerlim_radius, diagonal_length / 350)  # just some arbitrary scaling that looks nice
        colors = np.array([0, 0, 0] * positions.shape[0], dtype=np.float32)
        radii = np.array([radius] * positions.shape[0], dtype=np.float32)
        self.box[0] = self.renderer.draw_cylinders_from_to(
            positions,
            radii,
            colors,
            25,
        )
        self.update()

        self.main_window.update_action_texts()

    def identify_selected_sphere(self, xpos: int, ypos: int) -> int:
        """Return index of sphere that has been selected by clicking.

        :param xpos: x position of the mouse-click event
        :param ypos: y position of the mouse-click event
        """
        if len(self.structures) != 1:
            return -1

        click_position = np.array(
            [
                (xpos * 2 - self.width()) / self.width(),
                (ypos * 2 - self.height()) / self.height(),
            ],
            dtype=np.float32,
        )
        return select_sphere(
            click_position,
            self.camera.position,
            self.camera.view_matrix_inv,
            self.camera.projection_matrix_inv,
            self.camera.fov,
            self.height() / self.width(),
            self.structures[0].drawer.atom_positions,
            self.structures[0].drawer.atom_scales[:, 0],  # type: ignore[call-overload]
        )

    def exec_select_sphere(self, sphere_id: int, selected_spheres_list: list) -> None:
        """Select a sphere, change its color, update the selected spheres list.

        :param sphere_id: id of the sphere that shall be selected
        :param selected_spheres_list: list of selected spheres
        """
        id_in_selection = selected_spheres_list.index(-1)
        selected_spheres_list[id_in_selection] = sphere_id
        self.old_sphere_colors[id_in_selection] = self.structures[0].drawer.atom_colors[sphere_id].copy()
        self.structures[0].drawer.atom_colors[sphere_id] = self.new_sphere_colors[id_in_selection].copy()

    def exec_unselect_sphere(self, sphere_id: int, selected_spheres_list: list) -> None:
        """Unselect a sphere, change its color, update the selected spheres list.

        :param sphere_id: id of the sphere that shall be unselected
        :param selected_spheres_list: list of selected spheres
        """
        if sphere_id == -1:
            return

        id_in_selection = selected_spheres_list.index(sphere_id)
        self.structures[0].drawer.atom_colors[sphere_id] = self.old_sphere_colors[id_in_selection].copy()
        self.measurement_selected_spheres[id_in_selection] = -1

    def update_selected_atoms(self, purpose: int, event: QMouseEvent) -> None:
        """Update the selected atoms in the measurement or builder dialog.

        :param purpose: purpose of the selection (MEASUREMENT=0, BUILDER=1)
        :param event: mouse event (such as left click, right click...)
        :return:
        """
        if len(self.structures) != 1:
            return
        self.makeCurrent()
        selected_sphere_id = self.identify_selected_sphere(event.position().x(), event.position().y())

        selected_spheres_list = (
            self.measurement_selected_spheres if purpose == MEASUREMENT else self.builder_selected_spheres
        )

        if selected_sphere_id != -1:
            if -1 in selected_spheres_list:
                if selected_sphere_id in selected_spheres_list:
                    self.exec_unselect_sphere(selected_sphere_id, selected_spheres_list)
                else:
                    self.exec_select_sphere(selected_sphere_id, selected_spheres_list)
            elif selected_sphere_id in self.measurement_selected_spheres:
                self.exec_unselect_sphere(selected_sphere_id, selected_spheres_list)
        elif bool(QGuiApplication.keyboardModifiers() & Qt.ControlModifier):  # type: ignore[attr-defined]
            for selected_sphere_i in self.measurement_selected_spheres:
                self.exec_unselect_sphere(selected_sphere_i, selected_spheres_list)

        self.renderer.update_atoms_vao(
            self.structures[0].drawer.sphere.vertices,
            self.structures[0].drawer.sphere.indices,
            self.structures[0].drawer.sphere_model_matrices,
            self.structures[0].drawer.atom_colors,
        )
        self.update()

        self.main_window.measurement_dialog.display_metrics(
            self.structures[0],
            self.measurement_selected_spheres,
        ) if purpose == MEASUREMENT else None

    def reset_measurement(self) -> None:
        """Reset measurement arrays and measurement dialog."""
        if len(self.structures) != 1:
            return
        self.measurement_selected_spheres = [-1] * 4
        self.main_window.measurement_dialog.display_metrics(
            self.structures[0],
            self.measurement_selected_spheres,
        )

    def unselect_all_atoms(self) -> None:
        """Unselect all selected atoms."""
        if len(self.structures) != 1:
            return
        for selected_sphere_i in self.measurement_selected_spheres:
            if selected_sphere_i == -1:
                continue
            color = self.old_sphere_colors[self.measurement_selected_spheres.index(selected_sphere_i)].copy()
            self.structures[0].drawer.atom_colors[selected_sphere_i] = color
        for i in range(4):
            self.measurement_selected_spheres[i] = -1
        self.set_vertex_attribute_objects(update_bonds=False)
        self.update()

    def clear_builder_selected_atoms(self) -> None:
        """Reset the selected spheres builder spheres."""
        self.builder_selected_spheres = [-1] * 3

    def adopt_config(self, other_widget: StructureWidget, custom_geometry: tuple[int, int] | None = None) -> None:
        """Adopt the configuration of another StructureWidget object.

        :param other_widget: the other StructureWidget object
        :param custom_geometry: custom geometry (width, height) for the widget
        """
        geometry = self.geometry()
        if custom_geometry is None:
            other_geometry = other_widget.geometry()
            geometry.setWidth(other_geometry.width())
            geometry.setHeight(other_geometry.height())
        else:
            geometry.setWidth(custom_geometry[0])
            geometry.setHeight(custom_geometry[1])

        self.setGeometry(geometry)
        self.camera.adopt_config(other_widget.camera)
        self.set_structure(other_widget.structures, reset_view=False)
        # self.update()

        if other_widget.box[0] != -1:
            self.toggle_unit_cell_boundaries()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'crystalstructure_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QCheckBox, QComboBox,
    QDialog, QDialogButtonBox, QDoubleSpinBox, QFrame,
    QHeaderView, QLabel, QLineEdit, QPushButton,
    QSizePolicy, QSpinBox, QTableWidget, QTableWidgetItem,
    QWidget)

class Ui_CrystalDialog(object):
    def setupUi(self, CrystalDialog):
        if not CrystalDialog.objectName():
            CrystalDialog.setObjectName(u"CrystalDialog")
        CrystalDialog.resize(400, 476)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(CrystalDialog.sizePolicy().hasHeightForWidth())
        CrystalDialog.setSizePolicy(sizePolicy)
        self.buttonBox = QDialogButtonBox(CrystalDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setGeometry(QRect(50, 440, 341, 32))
        self.buttonBox.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)
        self.buttonAddAtom = QPushButton(CrystalDialog)
        self.buttonAddAtom.setObjectName(u"buttonAddAtom")
        self.buttonAddAtom.setGeometry(QRect(300, 70, 90, 28))
        self.buttonAddAtom.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.checkBoxPreview = QCheckBox(CrystalDialog)
        self.checkBoxPreview.setObjectName(u"checkBoxPreview")
        self.checkBoxPreview.setEnabled(False)
        self.checkBoxPreview.setGeometry(QRect(200, 410, 131, 21))
        self.checkBoxPreview.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.line = QFrame(CrystalDialog)
        self.line.setObjectName(u"line")
        self.line.setGeometry(QRect(110, 10, 20, 461))
        self.line.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.line.setFrameShape(QFrame.VLine)
        self.line.setFrameShadow(QFrame.Sunken)
        self.selectCrystalSystem = QComboBox(CrystalDialog)
        self.selectCrystalSystem.addItem("")
        self.selectCrystalSystem.addItem("")
        self.selectCrystalSystem.addItem("")
        self.selectCrystalSystem.addItem("")
        self.selectCrystalSystem.addItem("")
        self.selectCrystalSystem.addItem("")
        self.selectCrystalSystem.addItem("")
        self.selectCrystalSystem.setObjectName(u"selectCrystalSystem")
        self.selectCrystalSystem.setEnabled(True)
        self.selectCrystalSystem.setGeometry(QRect(20, 10, 76, 24))
        self.selectCrystalSystem.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.selectSpaceGroup = QComboBox(CrystalDialog)
        self.selectSpaceGroup.addItem("")
        self.selectSpaceGroup.addItem("")
        self.selectSpaceGroup.addItem("")
        self.selectSpaceGroup.addItem("")
        self.selectSpaceGroup.setObjectName(u"selectSpaceGroup")
        self.selectSpaceGroup.setEnabled(True)
        self.selectSpaceGroup.setGeometry(QRect(20, 40, 76, 24))
        self.selectSpaceGroup.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputLatConst_a = QDoubleSpinBox(CrystalDialog)
        self.inputLatConst_a.setObjectName(u"inputLatConst_a")
        self.inputLatConst_a.setGeometry(QRect(47, 110, 65, 26))
        self.inputLatConst_a.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputLatConst_a.setMinimum(0.010000000000000)
        self.inputLatConst_a.setSingleStep(0.010000000000000)
        self.inputLatConst_a.setValue(1.000000000000000)
        self.inputLatConst_b = QDoubleSpinBox(CrystalDialog)
        self.inputLatConst_b.setObjectName(u"inputLatConst_b")
        self.inputLatConst_b.setEnabled(False)
        self.inputLatConst_b.setGeometry(QRect(47, 140, 65, 26))
        self.inputLatConst_b.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputLatConst_b.setMinimum(0.010000000000000)
        self.inputLatConst_b.setSingleStep(0.010000000000000)
        self.inputLatConst_b.setValue(1.000000000000000)
        self.inputLatConst_c = QDoubleSpinBox(CrystalDialog)
        self.inputLatConst_c.setObjectName(u"inputLatConst_c")
        self.inputLatConst_c.setEnabled(False)
        self.inputLatConst_c.setGeometry(QRect(47, 170, 65, 26))
        self.inputLatConst_c.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputLatConst_c.setMinimum(0.010000000000000)
        self.inputLatConst_c.setSingleStep(0.010000000000000)
        self.inputLatConst_c.setValue(1.000000000000000)
        self.inputAtomCoord_a = QDoubleSpinBox(CrystalDialog)
        self.inputAtomCoord_a.setObjectName(u"inputAtomCoord_a")
        self.inputAtomCoord_a.setGeometry(QRect(190, 30, 62, 26))
        self.inputAtomCoord_a.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputAtomCoord_a.setMaximum(0.990000000000000)
        self.inputAtomCoord_a.setSingleStep(0.010000000000000)
        self.inputAtomCoord_b = QDoubleSpinBox(CrystalDialog)
        self.inputAtomCoord_b.setObjectName(u"inputAtomCoord_b")
        self.inputAtomCoord_b.setGeometry(QRect(260, 30, 62, 26))
        self.inputAtomCoord_b.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputAtomCoord_b.setMaximum(0.990000000000000)
        self.inputAtomCoord_b.setSingleStep(0.010000000000000)
        self.inputAtomCoord_c = QDoubleSpinBox(CrystalDialog)
        self.inputAtomCoord_c.setObjectName(u"inputAtomCoord_c")
        self.inputAtomCoord_c.setGeometry(QRect(330, 30, 62, 26))
        self.inputAtomCoord_c.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputAtomCoord_c.setMaximum(0.990000000000000)
        self.inputAtomCoord_c.setSingleStep(0.010000000000000)
        self.inputElementSymbol = QLineEdit(CrystalDialog)
        self.inputElementSymbol.setObjectName(u"inputElementSymbol")
        self.inputElementSymbol.setEnabled(True)
        self.inputElementSymbol.setGeometry(QRect(130, 30, 51, 28))
        self.inputElementSymbol.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelElementSymbol = QLabel(CrystalDialog)
        self.labelElementSymbol.setObjectName(u"labelElementSymbol")
        self.labelElementSymbol.setGeometry(QRect(130, 10, 58, 16))
        self.labelElementSymbol.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelCoord_a = QLabel(CrystalDialog)
        self.labelCoord_a.setObjectName(u"labelCoord_a")
        self.labelCoord_a.setGeometry(QRect(190, 10, 58, 16))
        self.labelCoord_a.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelCoord_b = QLabel(CrystalDialog)
        self.labelCoord_b.setObjectName(u"labelCoord_b")
        self.labelCoord_b.setGeometry(QRect(260, 10, 58, 16))
        self.labelCoord_b.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelCoord_c = QLabel(CrystalDialog)
        self.labelCoord_c.setObjectName(u"labelCoord_c")
        self.labelCoord_c.setGeometry(QRect(330, 10, 58, 16))
        self.labelCoord_c.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelLatConst_a = QLabel(CrystalDialog)
        self.labelLatConst_a.setObjectName(u"labelLatConst_a")
        self.labelLatConst_a.setGeometry(QRect(20, 110, 16, 20))
        self.labelLatConst_a.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelLatConst_b = QLabel(CrystalDialog)
        self.labelLatConst_b.setObjectName(u"labelLatConst_b")
        self.labelLatConst_b.setGeometry(QRect(20, 140, 16, 20))
        self.labelLatConst_b.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelLatConst_C = QLabel(CrystalDialog)
        self.labelLatConst_C.setObjectName(u"labelLatConst_C")
        self.labelLatConst_C.setGeometry(QRect(20, 170, 16, 20))
        self.labelLatConst_C.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelTitleLatConst = QLabel(CrystalDialog)
        self.labelTitleLatConst.setObjectName(u"labelTitleLatConst")
        self.labelTitleLatConst.setGeometry(QRect(10, 90, 101, 16))
        self.labelTitleLatConst.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputSupercell_a = QSpinBox(CrystalDialog)
        self.inputSupercell_a.setObjectName(u"inputSupercell_a")
        self.inputSupercell_a.setGeometry(QRect(50, 380, 42, 26))
        self.inputSupercell_a.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputSupercell_a.setMinimum(1)
        self.inputSupercell_b = QSpinBox(CrystalDialog)
        self.inputSupercell_b.setObjectName(u"inputSupercell_b")
        self.inputSupercell_b.setGeometry(QRect(50, 410, 42, 26))
        self.inputSupercell_b.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputSupercell_b.setMinimum(1)
        self.inputSupercell_c = QSpinBox(CrystalDialog)
        self.inputSupercell_c.setObjectName(u"inputSupercell_c")
        self.inputSupercell_c.setGeometry(QRect(50, 440, 42, 26))
        self.inputSupercell_c.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputSupercell_c.setMinimum(1)
        self.labelTitleSupercell = QLabel(CrystalDialog)
        self.labelTitleSupercell.setObjectName(u"labelTitleSupercell")
        self.labelTitleSupercell.setGeometry(QRect(10, 350, 101, 16))
        self.labelTitleSupercell.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelSupercell_a = QLabel(CrystalDialog)
        self.labelSupercell_a.setObjectName(u"labelSupercell_a")
        self.labelSupercell_a.setGeometry(QRect(20, 380, 16, 20))
        self.labelSupercell_a.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelSupercell_b = QLabel(CrystalDialog)
        self.labelSupercell_b.setObjectName(u"labelSupercell_b")
        self.labelSupercell_b.setGeometry(QRect(20, 410, 16, 20))
        self.labelSupercell_b.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelSupercell_c = QLabel(CrystalDialog)
        self.labelSupercell_c.setObjectName(u"labelSupercell_c")
        self.labelSupercell_c.setGeometry(QRect(20, 440, 16, 20))
        self.labelSupercell_c.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.listAtoms = QTableWidget(CrystalDialog)
        self.listAtoms.setObjectName(u"listAtoms")
        self.listAtoms.setGeometry(QRect(130, 110, 256, 291))
        self.listAtoms.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.listAtoms.horizontalHeader().setCascadingSectionResizes(False)
        self.listAtoms.horizontalHeader().setMinimumSectionSize(30)
        self.listAtoms.horizontalHeader().setDefaultSectionSize(60)
        self.listAtoms.horizontalHeader().setStretchLastSection(True)
        self.pushButton = QPushButton(CrystalDialog)
        self.pushButton.setObjectName(u"pushButton")
        self.pushButton.setGeometry(QRect(130, 70, 90, 28))
        self.pushButton.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelTitleLatAngles = QLabel(CrystalDialog)
        self.labelTitleLatAngles.setObjectName(u"labelTitleLatAngles")
        self.labelTitleLatAngles.setGeometry(QRect(10, 210, 101, 16))
        self.labelTitleLatAngles.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputLatAngle_beta = QDoubleSpinBox(CrystalDialog)
        self.inputLatAngle_beta.setObjectName(u"inputLatAngle_beta")
        self.inputLatAngle_beta.setEnabled(False)
        self.inputLatAngle_beta.setGeometry(QRect(47, 270, 65, 26))
        self.inputLatAngle_beta.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputLatAngle_beta.setMinimum(0.010000000000000)
        self.inputLatAngle_beta.setMaximum(179.990000000000009)
        self.inputLatAngle_beta.setSingleStep(0.010000000000000)
        self.inputLatAngle_beta.setValue(90.000000000000000)
        self.labelLatAngle_alpha = QLabel(CrystalDialog)
        self.labelLatAngle_alpha.setObjectName(u"labelLatAngle_alpha")
        self.labelLatAngle_alpha.setGeometry(QRect(20, 240, 16, 20))
        self.labelLatAngle_alpha.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.labelLatAngle_beta = QLabel(CrystalDialog)
        self.labelLatAngle_beta.setObjectName(u"labelLatAngle_beta")
        self.labelLatAngle_beta.setGeometry(QRect(20, 270, 16, 20))
        self.labelLatAngle_beta.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputLatAngle_alpha = QDoubleSpinBox(CrystalDialog)
        self.inputLatAngle_alpha.setObjectName(u"inputLatAngle_alpha")
        self.inputLatAngle_alpha.setGeometry(QRect(47, 240, 65, 26))
        self.inputLatAngle_alpha.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputLatAngle_alpha.setMinimum(0.010000000000000)
        self.inputLatAngle_alpha.setMaximum(179.990000000000009)
        self.inputLatAngle_alpha.setSingleStep(0.010000000000000)
        self.inputLatAngle_alpha.setValue(90.000000000000000)
        self.inputLatAngle_gamma = QDoubleSpinBox(CrystalDialog)
        self.inputLatAngle_gamma.setObjectName(u"inputLatAngle_gamma")
        self.inputLatAngle_gamma.setEnabled(False)
        self.inputLatAngle_gamma.setGeometry(QRect(47, 300, 65, 26))
        self.inputLatAngle_gamma.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        self.inputLatAngle_gamma.setMinimum(0.010000000000000)
        self.inputLatAngle_gamma.setMaximum(179.990000000000009)
        self.inputLatAngle_gamma.setSingleStep(0.010000000000000)
        self.inputLatAngle_gamma.setValue(90.000000000000000)
        self.labelLatAngle_gamma = QLabel(CrystalDialog)
        self.labelLatAngle_gamma.setObjectName(u"labelLatAngle_gamma")
        self.labelLatAngle_gamma.setGeometry(QRect(20, 300, 16, 20))
        self.labelLatAngle_gamma.setLocale(QLocale(QLocale.English, QLocale.UnitedKingdom))
        QWidget.setTabOrder(self.selectCrystalSystem, self.selectSpaceGroup)
        QWidget.setTabOrder(self.selectSpaceGroup, self.inputLatConst_a)
        QWidget.setTabOrder(self.inputLatConst_a, self.inputLatConst_b)
        QWidget.setTabOrder(self.inputLatConst_b, self.inputLatConst_c)
        QWidget.setTabOrder(self.inputLatConst_c, self.inputLatAngle_alpha)
        QWidget.setTabOrder(self.inputLatAngle_alpha, self.inputLatAngle_beta)
        QWidget.setTabOrder(self.inputLatAngle_beta, self.inputLatAngle_gamma)
        QWidget.setTabOrder(self.inputLatAngle_gamma, self.inputSupercell_a)
        QWidget.setTabOrder(self.inputSupercell_a, self.inputSupercell_b)
        QWidget.setTabOrder(self.inputSupercell_b, self.inputSupercell_c)
        QWidget.setTabOrder(self.inputSupercell_c, self.inputElementSymbol)
        QWidget.setTabOrder(self.inputElementSymbol, self.inputAtomCoord_a)
        QWidget.setTabOrder(self.inputAtomCoord_a, self.inputAtomCoord_b)
        QWidget.setTabOrder(self.inputAtomCoord_b, self.inputAtomCoord_c)
        QWidget.setTabOrder(self.inputAtomCoord_c, self.buttonAddAtom)
        QWidget.setTabOrder(self.buttonAddAtom, self.pushButton)
        QWidget.setTabOrder(self.pushButton, self.listAtoms)
        QWidget.setTabOrder(self.listAtoms, self.checkBoxPreview)

        self.retranslateUi(CrystalDialog)
        self.buttonBox.accepted.connect(CrystalDialog.accept)
        self.buttonBox.rejected.connect(CrystalDialog.reject)

        QMetaObject.connectSlotsByName(CrystalDialog)
    # setupUi

    def retranslateUi(self, CrystalDialog):
        CrystalDialog.setWindowTitle(QCoreApplication.translate("CrystalDialog", u"Create custom crystal structure", None))
        self.buttonAddAtom.setText(QCoreApplication.translate("CrystalDialog", u"Add Atom", None))
        self.checkBoxPreview.setText(QCoreApplication.translate("CrystalDialog", u"Show Preview", None))
        self.selectCrystalSystem.setItemText(0, QCoreApplication.translate("CrystalDialog", u"Cubic", None))
        self.selectCrystalSystem.setItemText(1, QCoreApplication.translate("CrystalDialog", u"Tetragonal", None))
        self.selectCrystalSystem.setItemText(2, QCoreApplication.translate("CrystalDialog", u"Orthorhombic", None))
        self.selectCrystalSystem.setItemText(3, QCoreApplication.translate("CrystalDialog", u"Hexagonal", None))
        self.selectCrystalSystem.setItemText(4, QCoreApplication.translate("CrystalDialog", u"Monoclinic", None))
        self.selectCrystalSystem.setItemText(5, QCoreApplication.translate("CrystalDialog", u"Triclinic", None))
        self.selectCrystalSystem.setItemText(6, QCoreApplication.translate("CrystalDialog", u"(Trigonal)", None))

        self.selectSpaceGroup.setItemText(0, QCoreApplication.translate("CrystalDialog", u"\u2013", None))
        self.selectSpaceGroup.setItemText(1, QCoreApplication.translate("CrystalDialog", u"m3m", None))
        self.selectSpaceGroup.setItemText(2, QCoreApplication.translate("CrystalDialog", u"4/mmm", None))
        self.selectSpaceGroup.setItemText(3, QCoreApplication.translate("CrystalDialog", u"mmm", None))

        self.labelElementSymbol.setText(QCoreApplication.translate("CrystalDialog", u"Element", None))
        self.labelCoord_a.setText(QCoreApplication.translate("CrystalDialog", u"coord. a", None))
        self.labelCoord_b.setText(QCoreApplication.translate("CrystalDialog", u"coord. b", None))
        self.labelCoord_c.setText(QCoreApplication.translate("CrystalDialog", u"coord. c", None))
        self.labelLatConst_a.setText(QCoreApplication.translate("CrystalDialog", u"a", None))
        self.labelLatConst_b.setText(QCoreApplication.translate("CrystalDialog", u"b", None))
        self.labelLatConst_C.setText(QCoreApplication.translate("CrystalDialog", u"c", None))
        self.labelTitleLatConst.setText(QCoreApplication.translate("CrystalDialog", u"Lattice constants", None))
        self.labelTitleSupercell.setText(QCoreApplication.translate("CrystalDialog", u"Supercell", None))
        self.labelSupercell_a.setText(QCoreApplication.translate("CrystalDialog", u"<html><head/><body><p>N<sub>a</sub><br/></p></body></html>", None))
        self.labelSupercell_b.setText(QCoreApplication.translate("CrystalDialog", u"<html><head/><body><p>N<sub>b</sub><br/></p></body></html>", None))
        self.labelSupercell_c.setText(QCoreApplication.translate("CrystalDialog", u"<html><head/><body><p>N<sub>c</sub><br/></p></body></html>", None))
        self.pushButton.setText(QCoreApplication.translate("CrystalDialog", u"Clear Atoms", None))
        self.labelTitleLatAngles.setText(QCoreApplication.translate("CrystalDialog", u"Lattice angles", None))
        self.labelLatAngle_alpha.setText(QCoreApplication.translate("CrystalDialog", u"\u03b1", None))
        self.labelLatAngle_beta.setText(QCoreApplication.translate("CrystalDialog", u"\u03b2", None))
        self.labelLatAngle_gamma.setText(QCoreApplication.translate("CrystalDialog", u"\u03b3", None))
    # retranslateUi



"""Exceptions for the structure module."""

from __future__ import annotations


class FileImporterError(Exception):
    """base class for errors occurring when loading molecules from file."""


class FileFormatError(FileImporterError):
    """raised when the file format is wrong or unsupported."""



"""Contains the Molecule class."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from molara.structure.mos import Mos
from molara.structure.structure import Structure

__copyright__ = "Copyright 2024, Molara"
if TYPE_CHECKING:
    from molara.structure.atom import Atom


class Molecule(Structure):
    """Creates a new Molecule object."""

    def __init__(
        self,
        atomic_numbers: np.ndarray,
        coordinates: np.ndarray,
        header: str | None = None,
        dummy: bool = False,
        draw_bonds: bool = True,
    ) -> None:
        """Create a new Molecule object.

        :param atomic_numbers:np.ndarray: atomic numbers of a atoms
        :param coordinates:np.ndarray: coordinates of the molecule
        :param header:str: header from the imported file
        :param dummy: bool: a dummy object.
        :param draw_bonds: bool: draw bonds between atoms (Default is True)
        """
        if dummy:
            self.dummy = True

        self.atomic_numbers = np.array(atomic_numbers)
        self.atoms: list[Atom] = []
        self.mos = Mos()
        self.vdw_rads: list[np.float32] = []
        self.subdivisions = 20
        self.gen_energy_information(header)
        self.aos: list = []
        super().__init__(atomic_numbers, coordinates, draw_bonds)

    def gen_energy_information(self, string: str | None) -> None:
        """Read the energy from the second line.

        :param string: file header from which energy info is extracted
        """
        self.energy = 0.0

        if isinstance(string, str):
            split_string = string.split()

            if "energy:" in split_string:
                index_e = split_string.index("energy:")

                if index_e + 1 < len(split_string):
                    self.energy = float(
                        string.split()[split_string.index("energy:") + 1],
                    )



"""Contains the rendering function for the opengl widget."""

# mypy: disable-error-code="name-defined"
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from OpenGL.GL import (
    GL_ARRAY_BUFFER,
    GL_COLOR_BUFFER_BIT,
    GL_DEPTH_BUFFER_BIT,
    GL_ELEMENT_ARRAY_BUFFER,
    GL_FALSE,
    GL_FILL,
    GL_FRONT_AND_BACK,
    GL_LINE,
    GL_POINTS,
    GL_TRIANGLES,
    GL_UNSIGNED_INT,
    GL_VERTEX_PROGRAM_POINT_SIZE,
    GLuint,
    glBindBuffer,
    glBindVertexArray,
    glClear,
    glDeleteBuffers,
    glDeleteVertexArrays,
    glDrawArrays,
    glDrawElementsInstanced,
    glEnable,
    glGetUniformLocation,
    glPolygonMode,
    glUniform1f,
    glUniform3fv,
    glUniformMatrix4fv,
    glUseProgram,
)

from molara.rendering.buffers import setup_vao, setup_vao_numbers
from molara.rendering.cylinder import Cylinder, calculate_cylinder_model_matrix
from molara.rendering.sphere import Sphere, calculate_sphere_model_matrix

if TYPE_CHECKING:
    from numpy import floating

    from molara.rendering.camera import Camera

__copyright__ = "Copyright 2024, Molara"


class Renderer:
    """Contains the rendering function for the opengl widget."""

    def __init__(self) -> None:
        """Create a Renderer object."""
        self.atoms_vao: dict = {"vao": 0, "n_atoms": 0, "n_vertices": 0, "buffers": []}
        self.bonds_vao: dict = {"vao": 0, "n_bonds": 0, "n_vertices": 0, "buffers": []}
        self.spheres: list[dict] = []
        self.aspect_ratio: float = 1.0
        self.cylinders: list[dict] = []
        self.number_vao: list[dict] = []
        self.shaders: list[GLuint] = [0]

    def set_shaders(self, shaders: list[GLuint]) -> None:
        """Set the shader program for the opengl widget.

        :param shaders: The shader programs of the opengl widget.
        """
        self.shaders = shaders

    def draw_cylinders(  # noqa: PLR0913
        self,
        positions: np.ndarray,
        directions: np.ndarray,
        radii: np.ndarray,
        lengths: np.ndarray,
        colors: np.ndarray,
        subdivisions: int,
    ) -> int:
        """Draws one or multiple cylinders.

        If only one cylinder is drawn, the positions, directions, radii, lengths and colors are given as np.ndarray
        containing only one array, for instance: positions = np.array([[0, 0, 0]]). If multiple cylinders are drawn,
        the positions, directions, radii, lengths and colors are given as np.ndarray containing multiple arrays, for
        instance: positions = np.array([[0, 0, 0], [1, 1, 1]]).

        :param positions: Positions of the cylinders.
        :type positions: numpy.array of numpy.float32
        :param directions: Directions of the cylinders.
        :type directions: numpy.array of numpy.float32
        :param radii: Radii of the cylinders.
        :type radii: numpy.array of numpy.float32
        :param lengths: Lengths of the cylinders.
        :type lengths: numpy.array of numpy.float32
        :param colors: Colors of the cylinders.
        :type colors: numpy.array of numpy.float32
        :param subdivisions: Number of subdivisions of the cylinder.
        :type subdivisions: int
        :return: Returns the index of the cylinder in the list of cylinders.
        """
        n_instances = len(positions)
        cylinder_mesh = Cylinder(subdivisions)

        model_matrices = np.array([])

        for i in range(n_instances):
            model_matrix = calculate_cylinder_model_matrix(
                np.array(positions[i], dtype=np.float32),
                float(radii[i]),
                float(lengths[i]),
                np.array(directions[i], dtype=np.float32),
            )
            model_matrices = model_matrix if i == 0 else np.concatenate((model_matrices, model_matrix))

        cylinder = {
            "vao": 0,
            "n_instances": n_instances,
            "n_vertices": len(cylinder_mesh.vertices),
            "buffers": [],
        }
        cylinder["vao"], cylinder["buffers"] = setup_vao(
            cylinder_mesh.vertices,
            cylinder_mesh.indices,
            model_matrices,
            colors,
        )

        # get index of new cylinder instances in list
        i_cylinder = -1

        if len(self.cylinders) == 0:
            i_cylinder = 0
            self.cylinders.append(cylinder)
            return i_cylinder

        for i, check_cylinder in enumerate(self.cylinders):
            if check_cylinder["vao"] == 0:
                i_cylinder = i
                self.cylinders[i_cylinder] = cylinder
                break

        if i_cylinder == -1:
            i_cylinder = len(self.cylinders)
            self.cylinders.append(cylinder)

        return i_cylinder

    def draw_cylinders_from_to(
        self,
        positions: np.ndarray,
        radii: np.ndarray,
        colors: np.ndarray,
        subdivisions: int,
    ) -> int:
        """Draws one or multiple cylinders.

        :param positions: Positions [[start, end], [start, end], ...] of the cylinders.
        :param radii: Radii of the cylinders.
        :param colors: Colors of the cylinders.
        :param subdivisions: Number of subdivisions of the cylinder.
        :return: Returns the index of the cylinder in the list of cylinders.
        """
        _directions: list[list[floating]] = []
        _lengths: list[floating] = []
        _positions_middle: list[list[list[floating]]] = []
        for pos12 in positions:
            pos1, pos2 = pos12
            _directions.append((pos2 - pos1).tolist())
            _lengths.append(np.linalg.norm(pos2 - pos1))
            _positions_middle.append((0.5 * (pos1 + pos2)).tolist())
        positions_middle = np.array(_positions_middle)
        lengths = np.array(_lengths)
        directions = np.array(_directions) / lengths[:, None]
        return self.draw_cylinders(positions_middle, -directions, radii, lengths, colors, subdivisions)

    def draw_spheres(
        self,
        positions: np.ndarray,
        radii: np.ndarray,
        colors: np.ndarray,
        subdivisions: int,
    ) -> int:
        """Draws one or multiple spheres.

        If only one sphere is drawn, the positions, radii and colors are given
        as np.ndarray containing only one array, for instance: positions = np.array([[0, 0, 0]]). If multiple
        spheres are drawn, the positions, radii and colors are given as np.ndarray containing multiple arrays, for
        instance: positions = np.array([[0, 0, 0], [1, 1, 1]]).

        :param positions: Positions of the spheres.
        :type positions: numpy.array of numpy.float32
        :param radii: Radii of the spheres.
        :type radii: numpy.array of numpy.float32
        :param colors: Colors of the spheres.
        :type colors: numpy.array of numpy.float32
        :param subdivisions: Number of subdivisions of the sphere.
        :type subdivisions: int
        :return: Returns the index of the sphere in the list of spheres.
        """
        n_instances = len(positions)
        sphere_mesh = Sphere(subdivisions)
        if n_instances == 1:
            model_matrices = calculate_sphere_model_matrix(positions[0], radii[0])
        else:
            for i in range(n_instances):
                model_matrix = calculate_sphere_model_matrix(positions[i], radii[i])
                model_matrices = model_matrix if i == 0 else np.concatenate((model_matrices, model_matrix))  # type: ignore[reportPossiblyUnboundVariable]

        sphere = {
            "vao": 0,
            "n_instances": n_instances,
            "n_vertices": len(sphere_mesh.vertices),
            "buffers": [],
        }
        sphere["vao"], sphere["buffers"] = setup_vao(
            sphere_mesh.vertices,
            sphere_mesh.indices,
            model_matrices,  # type: ignore[reportPossiblyUnboundVariable]
            colors,
        )

        # get index of new sphere instances in list
        i_sphere = -1
        if len(self.spheres) != 0:
            for i, check_sphere in enumerate(self.spheres):
                if check_sphere["vao"] == 0:
                    i_sphere = i
                    self.spheres[i_sphere] = sphere
            if i_sphere == -1:
                i_sphere = len(self.spheres)
                self.spheres.append(sphere)
        else:
            i_sphere = 0
            self.spheres.append(sphere)
        return i_sphere

    def remove_cylinder(self, i_cylinder: int) -> None:
        """Remove a cylinder from the list of cylinders.

        :param i_cylinder: Index of the cylinder to remove.
        :type i_cylinder: int
        :return:
        """
        if i_cylinder >= len(self.cylinders):
            return

        cylinder = self.cylinders[i_cylinder]
        if cylinder["vao"] != 0:
            glBindBuffer(GL_ARRAY_BUFFER, 0)
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)
            for buffer in cylinder["buffers"]:
                glDeleteBuffers(1, [buffer])
            glDeleteVertexArrays(1, [cylinder["vao"]])
        self.cylinders[i_cylinder] = {
            "vao": 0,
            "n_instances": 0,
            "n_vertices": 0,
            "buffers": [],
        }

    def remove_sphere(self, i_sphere: int) -> None:
        """Remove a sphere from the list of spheres.

        :param i_sphere: Index of the sphere to remove.
        :type i_sphere: int
        :return:
        """
        if i_sphere < len(self.spheres):
            sphere = self.spheres[i_sphere]
            if sphere["vao"] != 0:
                glBindBuffer(GL_ARRAY_BUFFER, 0)
                glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)
                for buffer in sphere["buffers"]:
                    glDeleteBuffers(1, [buffer])
                glDeleteVertexArrays(1, [sphere["vao"]])
            self.spheres[i_sphere] = {
                "vao": 0,
                "n_instances": 0,
                "n_vertices": 0,
                "buffers": [],
            }

    def update_atoms_vao(
        self,
        vertices: np.ndarray,
        indices: np.ndarray,
        model_matrices: np.ndarray,
        colors: np.ndarray,
    ) -> None:
        """Update the vertex attribute object for the atoms.

        :param vertices: Vertices in the following order x,y,z,nx,ny,nz,..., where xyz are the cartesian coordinates.
        :type vertices: numpy.array of numpy.float32
        :param indices: Gives the connectivity of the vertices.
        :type indices: numpy.array of numpy.uint32
        :param model_matrices: Each matrix gives the transformation from object space to world.
        :type model_matrices: numpy.array of numpy.float32
        :param colors: Colors of the atoms.
        :type colors: numpy.array of numpy.float32
        :return:
        """
        if self.atoms_vao["vao"] != 0:
            glBindBuffer(GL_ARRAY_BUFFER, 0)
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)
            for buffer in self.atoms_vao["buffers"]:
                glDeleteBuffers(1, int(buffer))
            glDeleteVertexArrays(1, int(self.atoms_vao["vao"]))
        self.atoms_vao["vao"], self.atoms_vao["buffers"] = setup_vao(
            vertices,
            indices,
            model_matrices,
            colors,
        )
        self.atoms_vao["n_atoms"] = len(model_matrices)
        self.atoms_vao["n_vertices"] = len(vertices)

    def draw_numbers(
        self,
        digits: np.ndarray,
        positions_3d: np.ndarray,
    ) -> None:
        """Update the vertex attribute object for the numbers.

        :param digits: Digits of the numbers.
        :param positions_3d: 3D positions of the numbers.
        :return:
        """
        if len(self.number_vao) != 0:
            for number in self.number_vao:
                if number["vao"] != 0:
                    glBindBuffer(GL_ARRAY_BUFFER, 0)
                    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)
                    for buffer in number["buffers"]:
                        glDeleteBuffers(1, buffer)
                    glDeleteVertexArrays(1, number["vao"])
        self.number_vao = []

        vao, buffers = setup_vao_numbers(digits, positions_3d)
        self.number_vao.append({"vao": vao, "n_instances": len(digits), "buffers": buffers})

    def update_bonds_vao(
        self,
        vertices: np.ndarray,
        indices: np.ndarray,
        model_matrices: np.ndarray,
        colors: np.ndarray,
    ) -> None:
        """Update the vertex attribute object for the bonds.

        :param vertices: Vertices in the following order x,y,z,nx,ny,nz,..., where xyz are the cartesian coordinates.
        :type vertices: numpy.array of numpy.float32
        :param indices: Gives the connectivity of the vertices.
        :type indices: numpy.array of numpy.uint32
        :param model_matrices: Each matrix gives the transformation from object space to world.
        :type model_matrices: numpy.array of numpy.float32
        :param colors: Colors of the bonds.
        :type colors: numpy.array of numpy.float32
        :return:
        """
        if self.bonds_vao["vao"] != 0:
            glBindBuffer(GL_ARRAY_BUFFER, 0)
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)
            for buffer in self.bonds_vao["buffers"]:
                glDeleteBuffers(1, int(buffer))
            glDeleteVertexArrays(1, int(self.bonds_vao["vao"]))
        self.bonds_vao["vao"], self.bonds_vao["buffers"] = setup_vao(
            vertices,
            indices,
            model_matrices,
            colors,
        )
        self.bonds_vao["n_bonds"] = len(model_matrices)
        self.bonds_vao["n_vertices"] = len(vertices)

    def draw_scene(
        self,
        camera: Camera,
        bonds: bool,
    ) -> None:
        """Draws the scene.

        :param camera: Camera object.
        :type camera: Camera
        :param bonds: If True, bonds are drawn.
        :type bonds: bool
        :return:
        """
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glUseProgram(self.shaders[0])
        light_direction_loc = glGetUniformLocation(self.shaders[0], "light_direction")
        proj_loc = glGetUniformLocation(self.shaders[0], "projection")
        camera_loc = glGetUniformLocation(self.shaders[0], "camera_position")
        view_loc = glGetUniformLocation(self.shaders[0], "view")

        light_direction = -camera.position - camera.up_vector * camera.distance_from_target * 0.5
        glUniform3fv(light_direction_loc, 1, light_direction)
        glUniform3fv(camera_loc, 1, camera.position)
        glUniformMatrix4fv(proj_loc, 1, GL_FALSE, camera.projection_matrix)
        glUniformMatrix4fv(view_loc, 1, GL_FALSE, camera.view_matrix)

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        # Draw atoms
        if self.atoms_vao["vao"] != 0:
            glBindVertexArray(self.atoms_vao["vao"])
            glDrawElementsInstanced(
                GL_TRIANGLES,
                self.atoms_vao["n_vertices"],
                GL_UNSIGNED_INT,
                None,
                self.atoms_vao["n_atoms"],
            )

        # Draw bonds
        if self.bonds_vao["vao"] != 0 and bonds:
            glBindVertexArray(self.bonds_vao["vao"])
            glDrawElementsInstanced(
                GL_TRIANGLES,
                self.bonds_vao["n_vertices"],
                GL_UNSIGNED_INT,
                None,
                self.bonds_vao["n_bonds"],
            )

        # Draw spheres
        for sphere in self.spheres:
            if sphere["vao"] != 0:
                glBindVertexArray(sphere["vao"])
                glDrawElementsInstanced(
                    GL_TRIANGLES,
                    sphere["n_vertices"],
                    GL_UNSIGNED_INT,
                    None,
                    sphere["n_instances"],
                )

        # Draw cylinders
        for cylinder in self.cylinders:
            if cylinder["vao"] != 0:
                glBindVertexArray(cylinder["vao"])
                glDrawElementsInstanced(
                    GL_TRIANGLES,
                    cylinder["n_vertices"],
                    GL_UNSIGNED_INT,
                    None,
                    cylinder["n_instances"],
                )
        glBindVertexArray(0)

    def display_numbers(self, camera: Camera, scale_factor: float) -> None:
        """Draws the lines."""
        number_scale = 0.25
        scale = number_scale * scale_factor

        glEnable(GL_VERTEX_PROGRAM_POINT_SIZE)
        glUseProgram(self.shaders[1])

        glBindVertexArray(self.number_vao[0]["vao"])
        # Uniform for aspect ratio
        aspect_ratio_location = glGetUniformLocation(self.shaders[1], "aspect_ratio")
        scale_loc = glGetUniformLocation(self.shaders[1], "scale")
        glUniform1f(scale_loc, scale)
        glUniform1f(aspect_ratio_location, self.aspect_ratio)

        proj_loc = glGetUniformLocation(self.shaders[1], "projection")
        view_loc = glGetUniformLocation(self.shaders[1], "view")

        glUniformMatrix4fv(proj_loc, 1, GL_FALSE, camera.projection_matrix)
        glUniformMatrix4fv(view_loc, 1, GL_FALSE, camera.view_matrix)

        # Uniform for color
        color_location = glGetUniformLocation(self.shaders[1], "color_in")
        glUniform3fv(color_location, 1, np.array([0.0, 0.0, 0.0], dtype=np.float32))

        # Draw instanced
        glDrawArrays(GL_POINTS, 0, self.number_vao[0]["n_instances"])

        glBindVertexArray(0)



"""Contains the Crystal class, which is a subclass of Structure."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy import constants

from molara.structure.atom import atomic_number_to_symbol, elements

from .structure import Structure

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

    try:
        from pymatgen.core import Structure as Pmgstructure
    except ImportError:
        pmgstructure = None
    try:
        from ase import Atoms
    except ImportError:
        Atoms = None

__copyright__ = "Copyright 2024, Molara"


dimension = range(1, 4)


class Crystal(Structure):
    """Class that represents a crystal supercell."""

    def __init__(
        self,
        atomic_nums: list[int],
        coords: list[list[float]],
        basis_vectors: list[list[float]] | ArrayLike,
        supercell_dims: list[int],
    ) -> None:
        """Create a crystal supercell based on given particle positions in unit cell and lattice basis vectors.

        Particle positions are given in terms of the basis vectors:
        E.g. the position (0.5, 0.5, 0.) is always the center of a unit cell wall, regardless of the crystal system.

        :param atomic_numbers: contains the atomic numbers of the particles specified for the unit cell.
        :param coordinates: Nx3 matrix of particle (fractional) coordinates in the unit cell,
            i.e., coordinates in terms of the basis vectors.
        :param basis_vectors: 3x3 matrix of the lattice basis vectors.
        :param supercell_dims: side lengths of the supercell in terms of the cell constants
        """
        self.atomic_nums_unitcell = atomic_nums
        self.elements = [atomic_number_to_symbol(i) for i in self.atomic_nums_unitcell]
        self.coords_unitcell = self._fold_coords_into_unitcell(coords)
        self.basis_vectors = basis_vectors
        self.energy = 0.0  # TD: implement energy calculation

        self.make_supercell(supercell_dims)
        self.molar_mass = np.sum([elements[i]["Atomic mass"] for i in self.elements])
        self.volume_unitcell = Crystal.calc_volume_unitcell(self.basis_vectors)
        self.density_unitcell = float((self.molar_mass / constants.Avogadro) / self.volume_unitcell * 1e24)

    def _fold_coords_into_unitcell(
        self,
        fractional_coords: ArrayLike,
    ) -> list[list[float]]:
        """Folds coordinates into unit cell.

        :param fractional_coords: particle positions in fractional coordinates
        """
        return np.mod(fractional_coords, 1.0).tolist()

    def make_supercell(self, supercell_dims: list[int]) -> None:
        """Create a supercell of the crystal.

        :param supercell_dims: side lengths of the supercell in terms of the cell constants
        """
        self.supercell_dims = supercell_dims
        steps_a = np.arange(supercell_dims[0])
        steps_b = np.arange(supercell_dims[1])
        steps_c = np.arange(supercell_dims[2])
        steps_a.shape = (*steps_a.shape, 1, 1)
        steps_b.shape = (1, *steps_b.shape, 1)
        steps_c.shape = (1, 1, *steps_c.shape)
        translations_a = 1.0 * steps_a + 0.0 * steps_b + 0.0 * steps_c
        translations_b = 0.0 * steps_a + 1.0 * steps_b + 0.0 * steps_c
        translations_c = 0.0 * steps_a + 0.0 * steps_b + 1.0 * steps_c
        translation_vectors = np.array(
            [
                translations_a.flatten(),
                translations_b.flatten(),
                translations_c.flatten(),
            ],
        ).T

        num_unit_cells = translation_vectors.shape[0]
        self.fractional_coords_supercell = np.empty((0, 3), float)
        self.atomic_nums_supercell = np.empty(0, int)

        # create supercell arrays
        for atomic_number_i, position_i in zip(
            self.atomic_nums_unitcell,
            self.coords_unitcell,
        ):
            self.atomic_nums_supercell = np.append(
                self.atomic_nums_supercell,
                [atomic_number_i] * num_unit_cells,
            )
            self.fractional_coords_supercell = np.append(
                self.fractional_coords_supercell,
                position_i + translation_vectors,
                axis=0,
            )

        # create extra atoms at edges of supercell (quasi periodic boundaries)
        extra_atomic_nums, extra_fractional_coords = self.make_supercell_edge_atoms()

        if extra_atomic_nums:
            self.atomic_nums_supercell = np.append(
                self.atomic_nums_supercell,
                extra_atomic_nums,
            )
            self.fractional_coords_supercell = np.append(
                self.fractional_coords_supercell,
                extra_fractional_coords,
                axis=0,
            )

        # transform fractional to cartesian coordinates and instantiate atoms in super().__init__
        self.cartesian_coordinates_supercell = Crystal.fractional_to_cartesian_coords(
            self.fractional_coords_supercell,
            self.basis_vectors,
        )
        # ... and instantiate atoms in super().__init__
        super().__init__(
            self.atomic_nums_supercell,
            self.cartesian_coordinates_supercell,
            draw_bonds=False,
        )

    @staticmethod
    def fractional_to_cartesian_coords(
        fractional_coords: ArrayLike,
        basis_vectors: ArrayLike,
    ) -> np.ndarray:
        """Transform fractional coordinates (coordinates in terms of basis vectors) to cartesian coordinates.

        :param fractional_coords: fractional coordinates of the atoms
        :param basis_vectors: basis vectors of the crystal lattice
        """
        return np.dot(fractional_coords, basis_vectors)

    def edge_atom_coords(self) -> tuple[np.ndarray, np.ndarray]:
        """Return the coordinates of the edge atoms in the supercell."""
        _fractional_coords_np = np.array(self.fractional_coords_supercell)
        if len(_fractional_coords_np.shape) != 2:  # noqa: PLR2004
            msg = "Faulty shape of fractional_coords_np array. Shape must be (N,3)."
            raise ValueError(msg)
        edges = np.where(_fractional_coords_np == 0)
        return edges[0], edges[1]

    def calc_number_of_supercell_atoms(self, supercell_dims: list[int]) -> int:
        """Calculate the number of atoms in the supercell."""
        num_atoms_unitcell = len(self.atomic_nums_unitcell)
        ids_edge_atoms = np.where(np.array(self.coords_unitcell) == 0)[0]
        ids_edge_atoms_unique = np.unique(ids_edge_atoms)
        num_edge_atoms_unitcell = len(ids_edge_atoms_unique)

        base_value = (num_atoms_unitcell - num_edge_atoms_unitcell) * np.prod(supercell_dims, dtype=int)
        for id_edge_atom in ids_edge_atoms_unique:
            factor = 1
            coords = np.array(self.coords_unitcell[id_edge_atom])
            factor *= supercell_dims[0] + 1 if coords[0] == 0.0 else supercell_dims[0]
            factor *= supercell_dims[1] + 1 if coords[1] == 0.0 else supercell_dims[1]
            factor *= supercell_dims[2] + 1 if coords[2] == 0.0 else supercell_dims[2]
            base_value += factor
        return base_value

    def make_supercell_edge_atoms(self) -> tuple[list[int], list[list[float]]]:
        """Extra atoms are created at supercell edges (periodic boundaries).

        :param atomic_nums: atomic numbers of the atoms
        :param fractional_coords: fractional coordinates,
          i.e., the coordinates of the atoms in terms of the basis vectors.
        :param supercell_dims: supercell dimensions, e.g., [3,2,5] for a 3x2x5 supercell.
        """
        # first get 2D info on where the zero-valued atom coordinates are located.
        # ids_edge_atoms contains the 0th-axis ids, i.e., the ids of the atoms with zero-valued coords.
        # ids_edge_atom_coords contains the 1st-axis ids, i.e., the ids of the specific coords that are zero-valued.
        # example: if there are five atoms, the 0th and the 3rd of which have coordinates (0, .5, .5) and (.5, 0, 0),
        # respectively, then ids_edge_atoms would be [0, 3, 3], and ids_edge_atom_coords would be [0, 1, 2].
        _fractional_coords_np = np.array(self.fractional_coords_supercell)
        _supercell_dims_np = np.array(self.supercell_dims)
        atomic_nums = self.atomic_nums_supercell

        ids_edge_atoms, ids_edge_atom_coords = self.edge_atom_coords()
        extra_atomic_nums = []  # atomic numbers of the newly created atoms
        extra_fractional_coords = []  # fractional coordinates of the newly created atoms

        # iterate over the relevant atoms
        for _id_atom_unique in np.unique(ids_edge_atoms):
            _atomic_num = atomic_nums[_id_atom_unique]
            # get the zero-valued coord ids of the respective atom
            _ids_atom_coords = ids_edge_atom_coords[ids_edge_atoms == _id_atom_unique]
            _fractional_coords_atom = _fractional_coords_np[_id_atom_unique]

            if len(_ids_atom_coords) == dimension[0]:  # e.g., (.5, 0, .5)
                extra_atomic_nums += [_atomic_num]
                extra_fractional_coords += [_fractional_coords_atom.copy()]
                id1 = _ids_atom_coords[0]
                dim1 = _supercell_dims_np[id1]
                extra_fractional_coords[-1][id1] = dim1

            elif len(_ids_atom_coords) == dimension[1]:  # e.g., (.5, 0, 0)
                extra_atomic_nums += [_atomic_num] * 3
                id1, id2 = _ids_atom_coords
                dim1, dim2 = _supercell_dims_np[_ids_atom_coords]
                extra_fractional_coords += [_fractional_coords_atom.copy()]
                extra_fractional_coords[-1][id1] = dim1  # (dim1,0)

                extra_fractional_coords += [extra_fractional_coords[-1].copy()]
                extra_fractional_coords[-1][id2] = dim2  # (dim1,dim2)

                extra_fractional_coords += [_fractional_coords_atom.copy()]
                extra_fractional_coords[-1][id2] = dim2  # (0,dim2)

            elif len(_ids_atom_coords) == dimension[2]:  # i.e., (0, 0, 0)
                extra_atomic_nums += [_atomic_num] * 7
                id1, id2, id3 = _ids_atom_coords
                dim1, dim2, dim3 = _supercell_dims_np[_ids_atom_coords]

                extra_fractional_coords += [_fractional_coords_atom.copy()]
                extra_fractional_coords[-1][id1] = dim1  # (dim1,0,0)
                extra_fractional_coords += [extra_fractional_coords[-1].copy()]
                extra_fractional_coords[-1][id2] = dim2  # (dim1,dim2,0)
                extra_fractional_coords += [extra_fractional_coords[-1].copy()]
                extra_fractional_coords[-1][id3] = dim3  # (dim1,dim2,dim3)

                extra_fractional_coords += [_fractional_coords_atom.copy()]
                extra_fractional_coords[-1][id1] = dim1
                extra_fractional_coords[-1][id3] = dim3  # (dim1,0,dim3)

                extra_fractional_coords += [_fractional_coords_atom.copy()]
                extra_fractional_coords[-1][id2] = dim2  # (0,dim2,0)
                extra_fractional_coords += [extra_fractional_coords[-1].copy()]
                extra_fractional_coords[-1][id3] = dim3  # (0,dim2,dim3)

                extra_fractional_coords += [_fractional_coords_atom.copy()]
                extra_fractional_coords[-1][id3] = dim3  # (0,0,dim3)
            else:
                msg = "Unexpected number of zero-valued coordinates encountered."
                raise ValueError(msg)
        return extra_atomic_nums, extra_fractional_coords

    @property
    def unitcell_boundaries_positions(self) -> np.ndarray:
        """Return the positions of the unit cell box."""
        basis_vectors_matrix = np.array(self.basis_vectors)
        zero_vec = np.array([0, 0, 0])

        return (
            np.array(
                [
                    [zero_vec, basis_vectors_matrix[0]],
                    [zero_vec, basis_vectors_matrix[1]],
                    [zero_vec, basis_vectors_matrix[2]],
                    [basis_vectors_matrix[0], basis_vectors_matrix[0] + basis_vectors_matrix[1]],
                    [basis_vectors_matrix[0], basis_vectors_matrix[0] + basis_vectors_matrix[2]],
                    [basis_vectors_matrix[1], basis_vectors_matrix[1] + basis_vectors_matrix[0]],
                    [basis_vectors_matrix[1], basis_vectors_matrix[1] + basis_vectors_matrix[2]],
                    [basis_vectors_matrix[2], basis_vectors_matrix[2] + basis_vectors_matrix[1]],
                    [basis_vectors_matrix[2], basis_vectors_matrix[2] + basis_vectors_matrix[0]],
                    [
                        basis_vectors_matrix[0] + basis_vectors_matrix[1],
                        basis_vectors_matrix[0] + basis_vectors_matrix[1] + basis_vectors_matrix[2],
                    ],
                    [
                        basis_vectors_matrix[0] + basis_vectors_matrix[2],
                        basis_vectors_matrix[0] + basis_vectors_matrix[1] + basis_vectors_matrix[2],
                    ],
                    [
                        basis_vectors_matrix[1] + basis_vectors_matrix[2],
                        basis_vectors_matrix[0] + basis_vectors_matrix[1] + basis_vectors_matrix[2],
                    ],
                ],
                dtype=np.float32,
            )
            - self.center
        )

    @staticmethod
    def calc_volume_unitcell(basis_vectors: list[list[float]] | ArrayLike) -> float:
        """Calculate unit cell volume based on given lattice basis vectors.

        :param volume: unit cell volume to be matched
        """
        basis_vectors = np.array(basis_vectors)
        if basis_vectors.shape != (3, 3):
            msg = "Faulty shape of basis_vectors array. Shape must be (3,3)."
            raise ValueError(msg)
        # result is rounded to 12 digits because the det function tends to give results like 63.99999999999998
        return round(np.abs(np.linalg.det(basis_vectors)), 12)

    @classmethod
    def from_pymatgen(
        cls: type[Crystal],
        structure: Pmgstructure,
        supercell_dims: list[int],
    ) -> Crystal:
        """Create a Crystal object from a pymatgen.Structure object.

        :param structure: pymatgen.Structure object
        """
        return cls(list(structure.atomic_numbers), structure.frac_coords, structure.lattice.matrix, supercell_dims)

    @classmethod
    def from_ase(cls: type[Crystal], atoms: Atoms) -> Crystal:
        """Create a Crystal object from an ase.Atoms object.

        :params atoms: ase.Atoms object
        """
        assert atoms.get_pbc().all(), (
            "You are attempting to create a crystal from a non-periodic ase.Atoms object. "
            "For non-periodic systems, use Molecule.from_ase(). "
            "Partially periodic systems are not supported yet."
        )
        return cls(
            atoms.get_atomic_numbers(),
            atoms.get_scaled_positions(),
            atoms.get_cell(),
            supercell_dims=[1, 1, 1],
        )

    def copy(self) -> Crystal:
        """Return a copy of the Crystal object."""
        return Crystal(
            self.atomic_nums_unitcell,
            self.coords_unitcell,
            self.basis_vectors,
            self.supercell_dims,
        )

    """ overloading operators """

    def __mul__(self, supercell_dims: list[int]) -> Crystal:
        """Multiply Crystal by a sequence.

        Current implementation: multiply Crystal by a sequence of three integers [M, N, K]
        to create MxNxK supercell

        :param supercell_dims:  side lengths of the supercell in terms of the cell constants
        """
        crystal_copy = self.copy()
        crystal_copy.make_supercell(supercell_dims)
        return crystal_copy

    def __rmul__(self, supercell_dims: list[int]) -> Crystal:
        """Multiply Crystal by a sequence.

        :param supercell_dims:  side lengths of the supercell in terms of the cell constants
        """
        return self.__mul__(supercell_dims)



"""Dialog for building molecules via a Z-matrix."""

from __future__ import annotations

import re
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Callable, TypeVar

import numpy as np
from PySide6.QtWidgets import QDialog, QTableWidgetItem

from molara.gui.ui_builder import Ui_builder
from molara.structure.atom import element_symbol_to_atomic_number
from molara.structure.molecule import Molecule
from molara.structure.molecules import Molecules

if TYPE_CHECKING:
    from PySide6.QtWidgets import QMainWindow
    from typing_extensions import ParamSpec

    from molara.gui.main_window import MainWindow
    from molara.gui.structure_widget import StructureWidget

    T = TypeVar("T")
    P = ParamSpec("P")

__copyright__ = "Copyright 2024, Molara"


def toggle_slot(func: Callable[..., Any]) -> Callable[..., Any]:
    """Toggles the disable slot variable to circumvent recursivity."""

    def wrapper(self: BuilderDialog, *args: P.args, **kwargs: P.kwargs) -> None:
        if self.disable_slot:
            return
        self.disable_slot = True
        func(self, *args, **kwargs)
        self.disable_slot = False

    return wrapper


class BuilderDialog(QDialog):
    """Dialog to ask for information to build molecules."""

    def __init__(self, parent: QMainWindow | None = None) -> None:
        """Initialize the ZMatBuilder dialog.

        :param parent: the parent widget (main window).
        """
        super().__init__(
            parent,
        )  # structure widget is passed as a parent
        self.ui = Ui_builder()
        self.ui.setupUi(self)
        self.ui.AddAtomButton.clicked.connect(self.select_add)
        self.ui.DeleteAtomButton.clicked.connect(self.delete_atom)
        self.ui.tableWidget.itemChanged.connect(self.adapt_z_matrix)

        self.ui.tableWidget.acceptDrops()

        self.main_window: MainWindow = self.parent()
        self.structure_widget: StructureWidget = self.parent().structure_widget

        self.main_window.mols = Molecules()
        self.z_matrix: list[dict] = []

        self.disable_slot: bool = False
        self.colliding_idx: int | None = None

        self._initialize_table()

    def _initialize_table(self) -> None:
        """Initialize the table of the builder."""
        self.ui.tableWidget.setRowCount(0)
        self.ui.tableWidget.setColumnCount(7)

        column_widths = [55, 40, 70, 40, 70, 40, 70]
        for col, width in enumerate(column_widths):
            self.ui.tableWidget.setColumnWidth(col, width)

        text_data = ["Element", "At. 1", "Distance[A]", "At. 2", "Angle[°]", "At. 3", "Dihedral[°]"]
        for column, text in enumerate(text_data):
            self.ui.tableWidget.setHorizontalHeaderItem(column, QTableWidgetItem(text))

    @toggle_slot
    def select_add(self) -> None:
        """Select the add procedure. Depends on the number of atoms in the molecule."""
        self.disable_slot = False
        num_atoms = 0 if self.main_window.mols.num_mols == 0 else self.main_window.mols.mols[0].n_at
        params, atom_ids = self._get_parameters(num_atoms)
        self.exec_add_atom(num_atoms, params, atom_ids)

    def exec_add_atom(self, count_atoms: int, params: tuple, atom_ids: list) -> None:
        """Execute the addition of an atom."""
        self.add_atom(count_atoms, params, atom_ids)
        mol = self.main_window.mols.mols[0]

        if not self.err and self.colliding_idx is None:
            self.ui.ErrorMessageBrowser.setText("")
            self.structure_widget.set_structure([mol])
            self.structure_widget.update()
            self.z_matrix.append({"parameter": params, "atom_ids": atom_ids})
            self._set_z_matrix_row(mol.n_at, mol.n_at - 1)
            self.structure_widget.clear_builder_selected_atoms()

    def delete_atom(self) -> None:
        """Delete an atom from the z-matrix visualization table and z_matrix itself."""
        index = self.ui.tableWidget.currentRow()

        mol = self.main_window.mols.mols[0]

        do_deletion = self._check_z_matrix_deletion(index)

        if not do_deletion:
            return

        error_msg = f"Atom {index+1} will be deleted."
        self.ui.ErrorMessageBrowser.setText(error_msg)
        self._delete_zmat_row(index, mol.n_at)
        self._delete_table_row(index)

        if mol.n_at <= 2:  # noqa: PLR2004
            mol.draw_bonds = False

        mol.remove_atom(index=index)

        self.structure_widget.delete_structure()

        if mol.n_at > 0:
            self.structure_widget.set_structure([mol])
        else:
            self.main_window.mols.remove_molecule(0)
            self.structure_widget.update()

    @toggle_slot
    def adapt_z_matrix(self, item: QTableWidgetItem) -> None:
        """Change the z-matrix in dependence of the visualization table.

        :param item: passed item from the visualization table
        """
        row = item.row()

        self.disable_slot = False
        self._clear_upper_table_entries(self.main_window.mols.mols[0].n_at)

        params = self._get_parameters_from_table(row)

        if params[0] is None:
            error_msg = "Incorrect input type."
            self.ui.ErrorMessageBrowser.setText(error_msg)
            self._set_z_matrix_row(self.main_window.mols.mols[0].n_at, row)
            return

        self.z_matrix_temp = deepcopy(self.z_matrix)
        self.z_matrix[row]["parameter"] = params
        mol_temp = self.main_window.mols.mols[0]
        self.main_window.mols.remove_molecule(0)

        if not self.z_matrix:
            return

        mol: Molecule = Molecule(np.array([1]), np.array([[0.0, 0.0, 0.0]]), None, dummy=True, draw_bonds=False)
        for i in range(len(self.z_matrix)):
            params = self.z_matrix[i]["parameter"]
            atom_ids = self.z_matrix[i]["atom_ids"]
            if i == 0:
                self.disable_slot = False
            self.add_atom(i, params, atom_ids)
            mol = self.main_window.mols.mols[0]

            if self.colliding_idx is not None:
                mol = mol_temp
                self.z_matrix = self.z_matrix_temp
                self.main_window.mols.mols[0] = mol_temp
                break

        self.structure_widget.delete_structure()
        self.structure_widget.set_structure([mol])
        self._update_z_matrix(mol.n_at)

    def _orth(self, vec: np.ndarray, unitvec: np.ndarray) -> np.ndarray:
        """Calculate a vector that is orthogonal to the two given vectors.

        :param vec: vector to orthogonalize
        :param unitvec: unitvector

        """
        return vec - np.dot(vec, unitvec) * unitvec

    def add_atom(self, count_atoms: int, params: tuple, atom_ids: list) -> None:
        """Add an atom to the molecule.

        :param count_atoms: number of atoms that have been added so far
        :param params: parameters that are passed for the atom (angles in rad)
        :param atom_ids: atom ids of the selected atoms
        """
        self.err = False
        element, dist = params[0], params[1]
        angle = params[2] if len(params) > 2 else None  # noqa: PLR2004
        # dihedral = params[3] if len(params) > 3 else None

        atomic_num = element_symbol_to_atomic_number(element)
        atomic_num_valid = self._check_element(atomic_num)
        dist_valid = self._check_value(dist) if count_atoms > 0 else True
        angle_valid = self._check_value(angle) if angle is not None else True
        values_valid = atomic_num_valid and dist_valid and angle_valid

        if not (atomic_num_valid and values_valid):
            return

        # check if atom is the very first to be added
        if count_atoms == 0:
            pos = self.calc_position_new_atom(count_atoms, params, atom_ids)
            self.main_window.mols.add_molecule(
                Molecule(np.array([atomic_num]), pos, draw_bonds=False),
            )
            return

        # starting from the third atom, atoms must have been selected in order to specify relative position
        atom_selection_check = self._check_selected_atoms(*atom_ids[:count_atoms]) if count_atoms >= 2 else True  # noqa: PLR2004
        if not atom_selection_check:
            return

        mol = self.main_window.mols.mols[0]
        pos = self.calc_position_new_atom(count_atoms, params, atom_ids)

        # check for collisions
        self.colliding_idx = mol.compute_collision(pos) if count_atoms >= 3 else None  # noqa: PLR2004
        if self.colliding_idx is not None:
            error_msg = f"The atom would collide with atom {self.colliding_idx+1}."
            self.ui.ErrorMessageBrowser.setText(error_msg)
            return

        mol.toggle_bonds() if count_atoms == 1 else None
        mol.add_atom(atomic_num, np.squeeze(pos))

    def calc_position_new_atom(self, count_atoms: int, params: tuple, atom_ids: list) -> np.ndarray:
        """Calculate the position of the new atom to be added.

        :param count_atoms: number of atoms that have been added so far
        :param params: parameters that are passed for the atom
        :param atom_ids: atom ids of the selected atoms
        """
        if count_atoms == 0:
            return np.zeros([1, 3])

        dist = params[1]
        angle = params[2] if len(params) > 2 else None  # noqa: PLR2004
        dihedral = params[3] if len(params) > 3 else None  # noqa: PLR2004
        # add second atom
        if count_atoms == 1:
            return np.array([0.0, 0.0, dist])

        mol = self.main_window.mols.mols[0]
        # add third atom
        if count_atoms == 2:  # noqa: PLR2004
            coord = np.array([dist * np.sin(angle), 0, dist * np.cos(angle)])
            coord[2] = (
                mol.atoms[atom_ids[0]].position[2] - coord[2]
                if atom_ids[0] == 1
                else mol.atoms[atom_ids[0]].position[2] + coord[2]
            )
            return coord

        # add nth atom
        at1_id, at2_id, at3_id = atom_ids
        vec1 = mol.atoms[at2_id].position - mol.atoms[at1_id].position
        length = np.linalg.norm(vec1)
        vec1 = vec1 / length if length > 0 else np.array([0, 0, 1.0])
        vec2 = self._orth(mol.atoms[at3_id].position - mol.atoms[at2_id].position, vec1)
        length = np.linalg.norm(vec2)
        vec2 = vec2 / length if length > 0 else np.array([1.0, 0, 0])
        if np.linalg.norm(self._orth(vec2, vec1)) == 0:
            vec2 = np.array([0, 1.0, 0])
        vec3 = np.cross(vec1, vec2)
        vec3 = vec3 / np.linalg.norm(vec3)
        tmp = dist * np.sin(angle)
        coord = mol.atoms[at1_id].position + dist * np.cos(angle) * vec1
        coord += tmp * np.cos(dihedral) * vec2 + tmp * np.sin(dihedral) * vec3
        return coord

    def _clear_all_text(self) -> None:
        """Clear all text in the zbuilder dialog."""
        for line_edit in [
            self.ui.Box_0Element,
            self.ui.Box_1BondDistance,
            self.ui.Box_2BondAngle,
            self.ui.Box_3DihedralAngle,
        ]:
            line_edit.clear()

    def _check_value(self, *args: float, threshold: float = 1e-8) -> bool:
        """Check whether values are larger than a threshold.

        :param args: Parameter to check whether threshold is reached.
        :param threshold: Threshold to be reached.
        """
        error_msg = "Parameter values are not valid."
        vals_above_threshold = True

        for arg in args:
            vals_above_threshold = arg > threshold
            if not (vals_above_threshold):
                self.ui.ErrorMessageBrowser.setText(error_msg)
                self.err = True
                break

        return vals_above_threshold

    def _check_selected_atoms(self, *selected_atoms: int) -> bool:
        """Check if all necessary atoms are selected.

        :param selected_atoms:int: Selected Atoms to check if not -1
        """
        error_msg = "Not enough atoms selected."
        all_selected = True
        for idx in selected_atoms:
            if idx == -1:
                all_selected = False
                self.ui.ErrorMessageBrowser.setText(error_msg)
                self.err = True
                break

        return all_selected

    def _check_element(self, atomic_num: int | None) -> bool:
        """Check if the element which should be added exists.

        :param atomic_num: atomic number (core charge) which is None if not an element.
        """
        error_msg = "This is not an element."
        is_element = True
        if atomic_num == 0:
            is_element = False
            self.ui.ErrorMessageBrowser.setText(error_msg)
            self.err = True

        return is_element

    @toggle_slot
    def _set_z_matrix_row(self, tot_row: int, row: int) -> None:
        """Set a z matrix row in the table in builder dialog.

        :param tot_row: Total number of rows
        :param row: The idx of the row to be changed
        """
        param_rows = [0, 2, 4, 6]
        atom_id_rows = [0, 1, 3, 5]

        self.ui.tableWidget.setRowCount(tot_row)

        for i, text in enumerate(self.z_matrix[row]["parameter"]):
            if text is not None:
                if i == 0:  # atom id
                    temp_text = text
                elif i == 1:  # distance
                    temp_text = f"{text:.2f}"
                else:  # angles (bond angle, dihedral angle)
                    temp_text = f"{np.rad2deg(text):.2f}"

                self.ui.tableWidget.setItem(row, param_rows[i], QTableWidgetItem(temp_text))
                if param_rows[i] != 0:
                    atom_id = self.z_matrix[row]["atom_ids"][i - 1]
                    self.ui.tableWidget.setItem(row, atom_id_rows[i], QTableWidgetItem(str(atom_id + 1)))

    def _update_z_matrix(self, tot_row: int) -> None:
        """Update the complete z-matrix table.

        :param tot_row: Total number of rows
        """
        self.ui.tableWidget.setRowCount(tot_row)
        for j in range(tot_row):
            self._set_z_matrix_row(tot_row, j)

    def _delete_table_row(self, idx: int) -> None:
        """Delete a row from the table.

        :param idx: Index of the row to be deleted
        """
        self.ui.tableWidget.removeRow(idx)

    def _get_parameters(self, num_atoms: int) -> tuple[tuple, list]:
        """Return the parameter from the input boxes.

        :param num_atoms: Current number of atoms of the molecule.
        """
        element: str = self.ui.Box_0Element.text().capitalize()
        dist: float = float(self.ui.Box_1BondDistance.text())
        angle: float = np.deg2rad(float(self.ui.Box_2BondAngle.text()))
        dihedral: float = np.deg2rad(float(self.ui.Box_3DihedralAngle.text()))
        atom_ids = self.structure_widget.builder_selected_spheres

        if num_atoms == 0:
            return (element, None), []
        if num_atoms == 1:
            return (element, dist), [0]
        if num_atoms == 2:  # noqa: PLR2004
            return (element, dist, angle), atom_ids
        return (
            element,
            dist,
            angle,
            dihedral,
        ), atom_ids

    def _get_parameters_from_table(self, row: int) -> tuple:
        """Return the parameter of a specified row in the table.

        :param row: Index of the row of interest.
        """
        param_type_validity = True

        if row >= 0:
            element = str(self.ui.tableWidget.item(row, 0).text().capitalize())
            param_type_validity = bool(re.match("^[A-Z]", element))
            param_type_validity = bool(element_symbol_to_atomic_number(element)) and param_type_validity

        if row >= 1:
            dist = self.ui.tableWidget.item(row, 2).text()
            param_type_validity = bool(re.match(r"^-?\d+(\.\d+)?$", dist)) and param_type_validity

        if row >= 2:  # noqa: PLR2004
            angle = self.ui.tableWidget.item(row, 4).text()
            param_type_validity = bool(re.match(r"^-?\d+(\.\d+)?$", angle)) and param_type_validity

        if row >= 3:  # noqa: PLR2004
            dihedral = self.ui.tableWidget.item(row, 6).text()
            param_type_validity = bool(re.match(r"^-?\d+(\.\d+)?$", dihedral)) and param_type_validity

        if not param_type_validity:
            return (None,)

        if row == 0:
            return (element, None)  # type: ignore[reportPossiblyUnboundVariable]
        if row == 1:
            return (element, float(dist))  # type: ignore[reportPossiblyUnboundVariable]
        if row == 2:  # noqa: PLR2004
            return (element, float(dist), np.deg2rad(float(angle)))  # type: ignore[reportPossiblyUnboundVariable]

        return (element, float(dist), np.deg2rad(float(angle)), np.deg2rad(float(dihedral)))  # type: ignore[reportPossiblyUnboundVariable]

    def _check_z_matrix_deletion(self, idx: int) -> bool:
        """Check if the deletion of the z-matrix entry is valid.

        :param idx: Index of the row to be deleted
        """
        if idx == -1:
            error_msg = "No Atom was chosen to be deleted."
            self.ui.ErrorMessageBrowser.setText(error_msg)
            return False

        for entry in self.z_matrix:
            if idx in entry["atom_ids"]:
                error_msg = f"Cannot be deleted. Atom {idx+1} depends on this atom."
                self.ui.ErrorMessageBrowser.setText(error_msg)
                return False

        return True

    def _delete_zmat_row(self, idx: int, num_atoms: int) -> None:
        """Delete an entry in the z-matrix.

        :param idx: Index of the row to be deleted
        :param num_atoms: Total number of atoms in the molecule.
        """
        for i in range(idx, num_atoms):
            for j, at_id in enumerate(self.z_matrix[i]["atom_ids"]):
                if at_id > idx and at_id != -1:
                    self.z_matrix[i]["atom_ids"][j] -= 1

        self.z_matrix.pop(idx)

    @toggle_slot
    def _clear_upper_table_entries(self, number_of_rows: int) -> None:
        """Clear the unnecessary table entries if there is some input."""
        first_row = 0
        second_row = 1
        third_row = 2
        if number_of_rows > first_row:
            for col in range(1, 6):
                self.ui.tableWidget.setItem(first_row, col, QTableWidgetItem(""))
        if number_of_rows > second_row:
            for col in range(3, 6):
                self.ui.tableWidget.setItem(second_row, col, QTableWidgetItem(""))
        if number_of_rows > third_row:
            for col in range(5, 6):
                self.ui.tableWidget.setItem(third_row, col, QTableWidgetItem(""))



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'structure_customizer.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDialog, QDoubleSpinBox,
    QLabel, QPushButton, QSizePolicy, QTextEdit,
    QWidget)

class Ui_structure_customizer(object):
    def setupUi(self, structure_customizer):
        if not structure_customizer.objectName():
            structure_customizer.setObjectName(u"structure_customizer")
        structure_customizer.resize(256, 191)
        structure_customizer.setMinimumSize(QSize(256, 152))
        structure_customizer.setMaximumSize(QSize(256, 191))
        self.viewModeButton = QPushButton(structure_customizer)
        self.viewModeButton.setObjectName(u"viewModeButton")
        self.viewModeButton.setGeometry(QRect(6, 5, 107, 32))
        self.viewModeButton.setAutoDefault(False)
        self.ballSizeSpinBox = QDoubleSpinBox(structure_customizer)
        self.ballSizeSpinBox.setObjectName(u"ballSizeSpinBox")
        self.ballSizeSpinBox.setGeometry(QRect(186, 10, 62, 22))
        self.ballSizeSpinBox.setMinimum(0.100000000000000)
        self.ballSizeSpinBox.setSingleStep(0.100000000000000)
        self.ballSizeSpinBox.setValue(1.000000000000000)
        self.stickSizeSpinBox = QDoubleSpinBox(structure_customizer)
        self.stickSizeSpinBox.setObjectName(u"stickSizeSpinBox")
        self.stickSizeSpinBox.setGeometry(QRect(186, 40, 62, 22))
        self.stickSizeSpinBox.setMinimum(0.100000000000000)
        self.stickSizeSpinBox.setSingleStep(0.100000000000000)
        self.ballSizeLabel = QLabel(structure_customizer)
        self.ballSizeLabel.setObjectName(u"ballSizeLabel")
        self.ballSizeLabel.setGeometry(QRect(118, 10, 71, 20))
        self.stickSizeLabel = QLabel(structure_customizer)
        self.stickSizeLabel.setObjectName(u"stickSizeLabel")
        self.stickSizeLabel.setGeometry(QRect(118, 40, 71, 20))
        self.toggleBondsButton = QPushButton(structure_customizer)
        self.toggleBondsButton.setObjectName(u"toggleBondsButton")
        self.toggleBondsButton.setGeometry(QRect(6, 35, 107, 32))
        self.toggleBondsButton.setAutoDefault(False)
        self.loadSelect = QComboBox(structure_customizer)
        self.loadSelect.setObjectName(u"loadSelect")
        self.loadSelect.setGeometry(QRect(1, 155, 141, 32))
        self.loadButton = QPushButton(structure_customizer)
        self.loadButton.setObjectName(u"loadButton")
        self.loadButton.setGeometry(QRect(6, 125, 61, 32))
        self.loadButton.setAutoDefault(False)
        self.saveButton = QPushButton(structure_customizer)
        self.saveButton.setObjectName(u"saveButton")
        self.saveButton.setGeometry(QRect(140, 125, 108, 32))
        self.saveButton.setAutoDefault(False)
        self.saveName = QTextEdit(structure_customizer)
        self.saveName.setObjectName(u"saveName")
        self.saveName.setGeometry(QRect(140, 159, 107, 21))
        self.deleteButton = QPushButton(structure_customizer)
        self.deleteButton.setObjectName(u"deleteButton")
        self.deleteButton.setGeometry(QRect(74, 125, 61, 32))
        self.deleteButton.setAutoDefault(False)
        self.toggleNumbersButton = QPushButton(structure_customizer)
        self.toggleNumbersButton.setObjectName(u"toggleNumbersButton")
        self.toggleNumbersButton.setGeometry(QRect(6, 65, 107, 32))
        self.toggleNumbersButton.setAutoDefault(False)
        self.indexSizeLabel = QLabel(structure_customizer)
        self.indexSizeLabel.setObjectName(u"indexSizeLabel")
        self.indexSizeLabel.setGeometry(QRect(118, 70, 71, 20))
        self.indexSizeSpinBox = QDoubleSpinBox(structure_customizer)
        self.indexSizeSpinBox.setObjectName(u"indexSizeSpinBox")
        self.indexSizeSpinBox.setGeometry(QRect(186, 70, 62, 22))
        self.indexSizeSpinBox.setMinimum(0.100000000000000)
        self.indexSizeSpinBox.setMaximum(3.000000000000000)
        self.indexSizeSpinBox.setSingleStep(0.050000000000000)
        self.indexSizeSpinBox.setValue(1.000000000000000)
        self.colorSchemeLabel = QLabel(structure_customizer)
        self.colorSchemeLabel.setObjectName(u"colorSchemeLabel")
        self.colorSchemeLabel.setGeometry(QRect(10, 100, 133, 16))
        self.colorSchemeSelect = QComboBox(structure_customizer)
        self.colorSchemeSelect.setObjectName(u"colorSchemeSelect")
        self.colorSchemeSelect.setGeometry(QRect(134, 95, 120, 32))

        self.retranslateUi(structure_customizer)

        QMetaObject.connectSlotsByName(structure_customizer)
    # setupUi

    def retranslateUi(self, structure_customizer):
        structure_customizer.setWindowTitle(QCoreApplication.translate("structure_customizer", u"Structure Customizer", None))
        self.viewModeButton.setText(QCoreApplication.translate("structure_customizer", u"Stick Mode", None))
        self.ballSizeLabel.setText(QCoreApplication.translate("structure_customizer", u"Atom Size:", None))
        self.stickSizeLabel.setText(QCoreApplication.translate("structure_customizer", u"Bond Size:", None))
        self.toggleBondsButton.setText(QCoreApplication.translate("structure_customizer", u"Bonds Off", None))
        self.loadButton.setText(QCoreApplication.translate("structure_customizer", u"Load", None))
        self.saveButton.setText(QCoreApplication.translate("structure_customizer", u"Save", None))
        self.saveName.setHtml(QCoreApplication.translate("structure_customizer", u"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><meta charset=\"utf-8\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"hr { height: 1px; border-width: 0; }\n"
"li.unchecked::marker { content: \"\\2610\"; }\n"
"li.checked::marker { content: \"\\2612\"; }\n"
"</style></head><body style=\" font-family:'.AppleSystemUIFont'; font-size:13pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>", None))
        self.deleteButton.setText(QCoreApplication.translate("structure_customizer", u"Delete", None))
        self.toggleNumbersButton.setText(QCoreApplication.translate("structure_customizer", u"Show Indices", None))
        self.indexSizeLabel.setText(QCoreApplication.translate("structure_customizer", u"Index Size:", None))
        self.colorSchemeLabel.setText(QCoreApplication.translate("structure_customizer", u"Select Color Scheme:", None))
    # retranslateUi
